# -*- coding: utf-8 -*-

## Copyright 2009-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 2009-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


"""
:mod:`ll.ul4c` provides templating for XML/HTML as well as any other text-based
format. A template defines placeholders for data output and basic logic (like
loops and conditional blocks), that define how the final rendered output will
look.

:mod:`ll.ul4c` compiles a template to an internal format, which makes it
possible to implement template renderers in multiple programming languages.
"""


__docformat__ = "reStructuredText"


import re, types, datetime, urllib.parse as urlparse, json, collections, locale, itertools, random, datetime, unicodedata

import antlr3

from ll import color, misc


# Regular expression used for splitting dates in isoformat
datesplitter = re.compile("[-T:.]")


def register(name):
	from ll import ul4on
	def registration(cls):
		ul4on.register("de.livinglogic.ul4." + name)(cls)
		cls.type = name
		return cls
	return registration


class Object:
	fields = {}

	def __getitem__(self, key):
		if key in self.fields:
			return getattr(self, key)
		raise KeyError(key)


###
### Location information
###

@register("location")
class Location(Object):
	"""
	A :class:`Location` object contains information about the location of a
	template tag.
	"""
	__slots__ = ("root", "source", "type", "starttag", "endtag", "startcode", "endcode")
	fields = {"root", "source", "type", "starttag", "endtag", "startcode", "endcode", "tag", "code"}

	def __init__(self, root=None, source=None, type=None, starttag=None, endtag=None, startcode=None, endcode=None):
		"""
		Create a new :class:`Location` object. The arguments have the following
		meaning:

			:var:`root`
				The :class:`Template` object

			:var:`source`
				The complete source string

			:var:`type`
				The tag type (i.e. ``"for"``, ``"if"``, etc. or ``None`` for
				literal text)

			:var:`starttag`
				The start position of the start delimiter.

			:var:`endtag`
				The end position of the end delimiter.

			:var:`startcode`
				The start position of the tag code.

			:var:`endcode`
				The end position of the tag code.
		"""
		self.root = root
		self.source = source
		self.type = type
		self.starttag = starttag
		self.endtag = endtag
		self.startcode = startcode
		self.endcode = endcode

	@property
	def code(self):
		return self.source[self.startcode:self.endcode]

	@property
	def tag(self):
		return self.source[self.starttag:self.endtag]

	def __repr__(self):
		return "<{}.{} {} at {:#x}>".format(self.__class__.__module__, self.__class__.__name__, self, id(self))

	def pos(self):
		lastlinefeed = self.source.rfind("\n", 0, self.starttag)
		if lastlinefeed >= 0:
			return (self.source.count("\n", 0, self.starttag)+1, self.starttag-lastlinefeed)
		else:
			return (1, self.starttag + 1)

	def __str__(self):
		(line, col) = self.pos()
		return "{!r} at {}:{} (line {}, col {})".format(self.tag, self.starttag, self.endtag, line, col)

	def ul4ondump(self, encoder):
		encoder.dump(self.root)
		encoder.dump(self.source)
		encoder.dump(self.type)
		encoder.dump(self.starttag)
		encoder.dump(self.endtag)
		encoder.dump(self.startcode)
		encoder.dump(self.endcode)

	def ul4onload(self, decoder):
		self.root = decoder.load()
		self.source = decoder.load()
		self.type = decoder.load()
		self.starttag = decoder.load()
		self.endtag = decoder.load()
		self.startcode = decoder.load()
		self.endcode = decoder.load()


###
### Exceptions
###

class Error(Exception):
	"""
	Exception class that wraps another exception and provides a location.
	"""
	def __init__(self, node):
		self.node = node

	def __repr__(self):
		return "<{}.{} in {} at {:#x}>".format(self.__class__.__module__, self.__class__.__name__, self.node, id(self))

	def __str__(self):
		if isinstance(self.node, (Template, TemplateClosure)):
			if self.node.name is not None:
				return "in template named {}".format(self.node.name)
			else:
				return "in unnamed template"
		else:
			return "in {!r} in tag {}".format(str(self.node), self.node.location)


class BlockError(Exception):
	"""
	Exception that is raised by the compiler when an illegal block structure is
	detected (e.g. an ``<?end if?>`` without a previous ``<?if?>``).
	"""

	def __init__(self, message):
		self.message = message

	def __str__(self):
		return self.message


###
### Exceptions used by the interpreted code for flow control
###

class BreakException(Exception):
	pass


class ContinueException(Exception):
	pass


class ReturnException(Exception):
	def __init__(self, value):
		self.value = value


###
### Various versions of undefined objects
###

class Undefined(object):
	def __bool__(self):
		return False

	def __iter__(self):
		raise TypeError("{!r} doesn't support iteration".format(self))

	def __len__(self):
		raise AttributeError("{!r} has no len()".format(self))

	def __getattr__(self, key):
		raise AttributeError("{!r} has no attribute {!r}".format(self, key))

	def __getitem__(self, key):
		raise TypeError("{!r} doesn't support indexing (key={!r})".format(self, key))


class UndefinedKey(Undefined):
	def __init__(self, key):
		self.__key = key

	def __repr__(self):
		return "undefined object for key {!r}".format(self.__key)


class UndefinedVariable(Undefined):
	def __init__(self, name):
		self.__name = name

	def __repr__(self):
		return "undefined variable {!r}".format(self.__name)


class UndefinedIndex(Undefined):
	def __init__(self, index):
		self.__index = index

	def __repr__(self):
		return "undefined object at index {!r}".format(self.__index)


###
### Helper functions
###


def handleeval(f):
	def wrapped(self, vars):
		try:
			return (yield from f(self, vars))
		except (BreakException, ContinueException, ReturnException) as ex:
			raise
		except Error as ex:
			if ex.node.location is not self.location:
				raise Error(self) from ex
			else:
				raise
		except Exception as ex:
			raise Error(self) from ex
	return wrapped


def _unpackvar(vars, name, value):
	if isinstance(name, str):
		vars[name] = value
	else:
		if len(name) > len(value):
			raise TypeError("too many values to unpack (expected {})".format(len(name)))
		elif len(name) < len(value):
			raise TypeError("need more than {} value{} to unpack)".format(len(values), "ss" if len(values) != 1 else ""))
		for (name, value) in zip(name, value):
			_unpackvar(vars, name, value)


def _str(obj=""):
	if obj is None:
		return ""
	elif isinstance(obj, Undefined):
		return ""
	else:
		return str(obj)


def _repr(obj):
	if isinstance(obj, str):
		return repr(obj)
	elif isinstance(obj, datetime.datetime):
		s = str(obj.isoformat())
		if s.endswith("T00:00:00"):
			s = s[:-9]
		return "@({})".format(s)
	elif isinstance(obj, datetime.date):
		return "@({})".format(obj.isoformat())
	elif isinstance(obj, datetime.timedelta):
		return repr(obj).partition(".")[-1]
	elif isinstance(obj, color.Color):
		if obj[3] == 0xff:
			s = "#{:02x}{:02x}{:02x}".format(obj[0], obj[1], obj[2])
			if s[1]==s[2] and s[3]==s[4] and s[5]==s[6]:
				return "#{}{}{}".format(s[1], s[3], s[5])
			return s
		else:
			s = "#{:02x}{:02x}{:02x}{:02x}".format(*obj)
			if s[1]==s[2] and s[3]==s[4] and s[5]==s[6] and s[7]==s[8]:
				return "#{}{}{}{}".format(s[1], s[3], s[5], s[7])
			return s
	elif isinstance(obj, collections.Sequence):
		return "[{}]".format(", ".join(_repr(item) for item in obj))
	elif isinstance(obj, collections.Mapping):
		return "{{{}}}".format(", ".join("{}: {}".format(_repr(key), _repr(value)) for (key, value) in obj.items()))
	else:
		return repr(obj)


def _asjson(obj):
	if obj is None:
		return "null"
	elif isinstance(obj, Undefined):
		return "{}.undefined"
	if isinstance(obj, (bool, int, float, str)):
		return json.dumps(obj)
	elif isinstance(obj, datetime.datetime):
		return "new Date({}, {}, {}, {}, {}, {}, {})".format(obj.year, obj.month-1, obj.day, obj.hour, obj.minute, obj.second, obj.microsecond//1000)
	elif isinstance(obj, datetime.date):
		return "new Date({}, {}, {})".format(obj.year, obj.month-1, obj.day)
	elif isinstance(obj, datetime.timedelta):
		return "ul4.TimeDelta.create({}, {}, {})".format(obj.days, obj.seconds, obj.microseconds)
	elif isinstance(obj, misc.monthdelta):
		return "ul4.MonthDelta.create({})".format(obj.months)
	elif isinstance(obj, color.Color):
		return "ul4.Color.create({}, {}, {}, {})".format(*obj)
	elif isinstance(obj, collections.Mapping):
		return "{{{}}}".format(", ".join("{}: {}".format(_asjson(key), _asjson(value)) for (key, value) in obj.items()))
	elif isinstance(obj, collections.Sequence):
		return "[{}]".format(", ".join(_asjson(item) for item in obj))
	elif isinstance(obj, Template):
		return obj.jssource()
	else:
		raise TypeError("can't handle object of type {}".format(type(obj)))


def _xmlescape(obj):
	if obj is None:
		return ""
	elif isinstance(obj, Undefined):
		return ""
	else:
		return misc.xmlescape(str(obj))


###
### Compiler stuff: Nodes for the AST
###

class AST(Object):
	"""
	Base class for all syntax tree nodes.
	"""

	# Set of attributes available via :meth:`getitem`.
	fields = {"type", "location", "start", "end"}

	# "Global" functions and methods. Functions in ``functions`` will be exposed to UL4 code
	functions = {}
	methods = {}

	def __init__(self, location=None, start=None, end=None):
		self.location = location
		self.start = start
		self.end = end

	def __getitem__(self, key):
		if key in self.fields:
			return getattr(self, key)
		raise KeyError(key)

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		p.text(repr(self))

	def __str__(self):
		v = []
		level = 0
		needlf = False
		for code in self._str():
			if code is None:
				needlf = True
			elif isinstance(code, int):
				level += code
			else:
				if needlf:
					v.append("\n")
					v.append(level*"\t")
					needlf = False
				v.append(code)
		if needlf:
			v.append("\n")
		return "".join(v)

	def _str(self):
		# Format :var:`self`.
		# This is used by :meth:`__str__.
		# ``_str`` is a generator an may output:
		# ``None``, which means: "add a line feed and an indentation here"
		# an int, which means: add the int to the indentation level
		# a string, which means: add this string to the output
		yield self.location.source[self.start:self.end].replace("\r\n", " ").replace("\n", " ")

	@misc.notimplemented
	def eval(self, vars):
		"""
		This evaluates the node.

		This is a generator, which yields the text output of the node. If the
		node returns a value (as most nodes do), this is done as the value of a
		:exc:`StopIteration` exception.
		"""

	def ul4ondump(self, encoder):
		encoder.dump(self.location)
		encoder.dump(self.start)
		encoder.dump(self.end)

	def ul4onload(self, decoder):
		self.location = decoder.load()
		self.start = decoder.load()
		self.end = decoder.load()

	@classmethod
	def makefunction(cls, f):
		name = f.__name__
		if name.startswith("function_"):
			name = name[9:]
		cls.functions[name] = f
		return f

	@classmethod
	def makemethod(cls, f):
		name = f.__name__
		if name.startswith("method_"):
			name = name[7:]
		cls.methods[name] = f
		return f


@register("text")
class Text(AST):
	"""
	AST node for literal text.
	"""

	def text(self):
		# If ``keepws`` is true, we output the literal text from the location info.
		# Otherwise we have to strip linefeeds and indentation
		text = self.location.code
		if not self.location.root.keepws:
			text = "".join(line.lstrip() for line in text.splitlines())
		return text

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {1!r} at {2:#x}>".format(self, self.text(), id(self))

	def _str(self):
		yield "text {!r}".format(self.text())

	def eval(self, vars):
		yield self.text()


@register("const")
class Const(AST):
	"""
	Load a constant
	"""
	fields = AST.fields.union({"value"})

	def __init__(self, location=None, start=None, end=None, value=None):
		super().__init__(location, start, end)
		self.value = value

	def eval(self, vars):
		yield from ()
		return self.value

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.value)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.value = decoder.load()

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.value!r} at {1:#x}>".format(self, id(self))


@register("list")
class List(AST):
	"""
	AST nodes for loading a list object.
	"""

	fields = AST.fields.union({"items"})

	def __init__(self, location=None, start=None, end=None, *items):
		super().__init__(location, start, end)
		self.items = list(items)

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.items!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				for (i, item) in enumerate(self.items):
					if i:
						p.breakable()
					else:
						p.breakable("")
					p.pretty(item)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		result = []
		for item in self.items:
			item = (yield from item.eval(vars))
			result.append(item)
		return result

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.items)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.items = decoder.load()


@register("listcomp")
class ListComp(AST):
	"""
	AST node for list comprehension.
	"""

	fields = AST.fields.union({"item", "varname", "container", "condition"})


	def __init__(self, location=None, start=None, end=None, item=None, varname=None, container=None, condition=None):
		super().__init__(location, start, end)
		self.item = item
		self.varname = varname
		self.container = container
		self.condition = condition

	def __repr__(self):
		s = "<{0.__class__.__module__}.{0.__class__.__qualname__} item={0.item!r} varname={0.varname!r} container={0.container!r}".format(self)
		if self.condition is not None:
			s += " condition={0.condition!r}".format(self)
		return s + " at {:#x}>".format(id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("{0.__class__.__module__}.{0.__class__.__qualname__}(...)".format(self))
		else:
			with p.group(4, "{0.__class__.__module__}.{0.__class__.__qualname__}(".format(self), ")"):
				p.breakable("")
				p.text("item=")
				p.pretty(self.item)
				p.text(",")
				p.breakable()
				p.text("varname=")
				p.pretty(self.varname)
				p.text(",")
				p.breakable()
				p.text("container=")
				p.pretty(self.container)
				if self.condition is not None:
					p.text(",")
					p.breakable()
					p.text("condition=")
					p.pretty(self.condition)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		container = (yield from self.container.eval(vars))
		vars = collections.ChainMap({}, vars) # Don't let loop variables leak into the surrounding scope
		result = []
		for item in container:
			_unpackvar(vars, self.varname, item)
			if self.condition is None or (yield from self.condition.eval(vars)):
				item = (yield from self.item.eval(vars))
				result.append(item)
		return result

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.item)
		encoder.dump(self.varname)
		encoder.dump(self.container)
		encoder.dump(self.condition)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.item = decoder.load()
		self.varname = decoder.load()
		self.container = decoder.load()
		self.condition = decoder.load()


@register("dict")
class Dict(AST):
	"""
	AST node for loading a dict object.
	"""

	fields = AST.fields.union({"items"})

	def __init__(self, location=None, start=None, end=None, *items):
		super().__init__(location, start, end)
		self.items = list(items)

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.items!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				for item in self.items:
					p.breakable()
					p.pretty(item[0])
					p.text("=")
					p.pretty(item[1])
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		result = {}
		for item in self.items:
			key = (yield from item[0].eval(vars))
			value = (yield from item[1].eval(vars))
			result[key] = value
		return result

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.items)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.items = [tuple(item) for item in decoder.load()]


@register("dictcomp")
class DictComp(AST):
	"""
	AST node for dictionary comprehension.
	"""

	fields = AST.fields.union({"key", "value", "varname", "container", "condition"})

	def __init__(self, location=None, start=None, end=None, key=None, value=None, varname=None, container=None, condition=None):
		super().__init__(location, start, end)
		self.key = key
		self.value = value
		self.varname = varname
		self.container = container
		self.condition = condition

	def __repr__(self):
		s = "<{0.__class__.__module__}.{0.__class__.__qualname__} key={0.key!r} value={0.value!r} varname={0.varname!r} container={0.container!r}".format(self)
		if self.condition is not None:
			s += " {0.condition!r}".format(self)
		return s + " at {:#x}>".format(id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("key=")
				p.pretty(self.key)
				p.breakable()
				p.text("value=")
				p.pretty(self.value)
				p.breakable()
				p.text("varname=")
				p.pretty(self.varname)
				p.breakable()
				p.text("container=")
				p.pretty(self.container)
				if self.condition is not None:
					p.breakable()
					p.text("condition=")
					p.pretty(self.condition)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		container = (yield from self.container.eval(vars))
		vars = collections.ChainMap({}, vars) # Don't let loop variables leak into the surrounding scope
		result = {}
		for item in container:
			_unpackvar(vars, self.varname, item)
			if self.condition is None or (yield from self.condition.eval(vars)):
				key = (yield from self.key.eval(vars))
				value = (yield from self.value.eval(vars))
				result[key] = value
		return result

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.key)
		encoder.dump(self.value)
		encoder.dump(self.varname)
		encoder.dump(self.container)
		encoder.dump(self.condition)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.key = decoder.load()
		self.value = decoder.load()
		self.varname = decoder.load()
		self.container = decoder.load()
		self.condition = decoder.load()


@register("genexpr")
class GenExpr(AST):
	"""
	AST node for a generator expression.
	"""

	fields = AST.fields.union({"item", "varname", "container", "condition"})

	def __init__(self, location=None, start=None, end=None, item=None, varname=None, container=None, condition=None):
		super().__init__(location, start, end)
		self.item = item
		self.varname = varname
		self.container = container
		self.condition = condition

	def __repr__(self):
		s = "<{0.__class__.__module__}.{0.__class__.__qualname__} item={0.item!r} varname={0.varname!r} container={0.container!r}".format(self)
		if self.condition is not None:
			s += " condition={0.condition!r}".format(self)
		return s + " at {:#x}>".format(id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("item=")
				p.pretty(self.item)
				p.breakable()
				p.text("varname=")
				p.pretty(self.varname)
				p.breakable()
				p.text("container=")
				p.pretty(self.container)
				if self.condition is not None:
					p.breakable()
					p.text("condition=")
					p.pretty(self.condition)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		container = (yield from self.container.eval(vars))
		vars = collections.ChainMap({}, vars) # Don't let loop variables leak into the surrounding scope
		def result():
			for item in container:
				_unpackvar(vars, self.varname, item)
				if self.condition is None or (yield from self.condition.eval(vars)):
					item = (yield from self.item.eval(vars))
					yield item
		return result()

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.item)
		encoder.dump(self.varname)
		encoder.dump(self.container)
		encoder.dump(self.condition)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.item = decoder.load()
		self.varname = decoder.load()
		self.container = decoder.load()
		self.condition = decoder.load()


@register("var")
class Var(AST):
	"""
	AST nodes for loading a variable.
	"""

	fields = AST.fields.union({"name"})

	def __init__(self, location=None, start=None, end=None, name=None):
		super().__init__(location, start, end)
		self.name = name

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.name!r} at {1:#x}>".format(self, id(self))

	@handleeval
	def eval(self, vars):
		yield from ()
		try:
			return vars[self.name]
		except KeyError:
			try:
				return self.functions[self.name]
			except KeyError:
				return UndefinedVariable(self.name)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.name)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.name = decoder.load()


class Block(AST):
	"""
	Base class for all AST nodes that are blocks.

	A block contains a sequence of AST nodes that are executed sequencially.
	A block may execute its content zero (e.g. an ``<?if?>`` block) or more times
	(e.g. a ``<?for?>`` block).
	"""

	fields = AST.fields.union({"endlocation", "content"})

	def __init__(self, location=None, start=None, end=None):
		super().__init__(location, start, end)
		self.endlocation = None
		self.content = []

	def append(self, item):
		self.content.append(item)

	def _str(self):
		if self.content:
			for node in self.content:
				yield from node._str()
				yield None
		else:
			yield "pass"
			yield None

	@handleeval
	def eval(self, vars):
		for node in self.content:
			yield from node.eval(vars)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.endlocation)
		encoder.dump(self.content)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.endlocation = decoder.load()
		self.content = decoder.load()


@register("ieie")
class IfElIfElse(Block):
	"""
	AST node for an conditional block.

	The content of the :class:`IfElIfElse` block is one :class:`If` block
	followed by zero or more :class:`ElIf` blocks followed by zero or one
	:class:`Else` block.
	"""
	def __init__(self, location=None, start=None, end=None, condition=None):
		super().__init__(location, start, end)
		if condition is not None:
			self.newblock(If(location, start, end, condition))

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {1} at {2:#x}>".format(self, repr(self.content)[1:-1], id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def append(self, item):
		self.content[-1].append(item)

	def newblock(self, block):
		if self.content:
			self.content[-1].endlocation = block.location
		self.content.append(block)

	def _str(self):
		for node in self.content:
			yield from node._str()

	@handleeval
	def eval(self, vars):
		for node in self.content:
			if isinstance(node, Else) or (yield from node.condition.eval(vars)):
				yield from node.eval(vars)
				break


@register("if")
class If(Block):
	"""
	AST node for an ``<?if?>`` block.
	"""

	fields = Block.fields.union({"condition"})

	def __init__(self, location=None, start=None, end=None, condition=None):
		super().__init__(location, start, end)
		self.condition = condition

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} condition={0.condition!r} {1} at {2:#x}>".format(self, " ..." if self.content else "", id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("condition=")
				p.pretty(self.condition)
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def _str(self):
		yield "if "
		yield from AST._str(self)
		yield ":"
		yield None
		yield +1
		yield from super()._str()
		yield -1

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.condition)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.condition = decoder.load()


@register("elif")
class ElIf(Block):
	"""
	AST node for an ``<?elif?>`` block.
	"""

	fields = Block.fields.union({"condition"})

	def __init__(self, location=None, start=None, end=None, condition=None):
		super().__init__(location, start, end)
		self.condition = condition

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} condition={0.condition!r} {1} at {2:#x}>".format(self, " ..." if self.content else "", id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("condition=")
				p.pretty(self.condition)
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def _str(self):
		yield "elif "
		yield from AST._str(self)
		yield ":"
		yield None
		yield +1
		yield from super()._str()
		yield -1

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.condition)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.condition = decoder.load()


@register("else")
class Else(Block):
	"""
	AST node for an ``<?else?>`` block.
	"""

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def _str(self):
		yield "else:"
		yield None
		yield +1
		yield from super()._str()
		yield -1


@register("for")
class For(Block):
	"""
	AST node for a ``<?for?>`` loop variable.
	"""

	fields = Block.fields.union({"varname", "container"})

	def __init__(self, location=None, start=None, end=None, varname=None, container=None):
		super().__init__(location, start, end)
		self.varname = varname
		self.container = container

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} varname={0.varname!r} container={0.container!r} {1} at {2:#x}>".format(self, " ..." if self.content else "", id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("varname=")
				p.pretty(self.varname)
				p.breakable()
				p.text("container=")
				p.pretty(self.container)
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.varname)
		encoder.dump(self.container)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.varname = decoder.load()
		self.container = decoder.load()

	def _str(self):
		yield "for "
		yield from AST._str(self)
		yield ":"
		yield None
		yield +1
		yield from super()._str()
		yield -1

	@handleeval
	def eval(self, vars):
		container = (yield from self.container.eval(vars))
		vars = collections.ChainMap({}, vars) # Don't let loop variables leak into the surrounding scope
		for item in container:
			_unpackvar(vars, self.varname, item)
			try:
				yield from super().eval(vars)
			except BreakException:
				break
			except ContinueException:
				pass


@register("break")
class Break(AST):
	"""
	AST node for a ``<?break?>`` inside a ``<?for?>`` block.
	"""

	def _str(self):
		yield "break"

	def eval(self, vars):
		yield from ()
		raise BreakException()


@register("continue")
class Continue(AST):
	"""
	AST node for a ``<?continue?>`` inside a ``<?for?>`` block.
	"""

	def _str(self):
		yield "continue"

	def eval(self, vars):
		yield from ()
		raise ContinueException()


@register("getattr")
class GetAttr(AST):
	"""
	AST node for getting an attribute from an object.

	The object is loaded from the AST node :var:`obj` and the attribute name
	is stored in the string :var:`attrname`.
	"""
	fields = AST.fields.union({"obj", "attrname"})

	def __init__(self, location=None, start=None, end=None, obj=None, attrname=None):
		super().__init__(location, start, end)
		self.obj = obj
		self.attrname = attrname

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} obj={0.obj!r}, attrname={0.attrname!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("obj=")
				p.pretty(self.obj)
				p.breakable()
				p.text("attrname=")
				p.pretty(self.attrname)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		obj = (yield from self.obj.eval(vars))
		try:
			return obj[self.attrname]
		except KeyError:
			return UndefinedKey(self.attrname)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)
		encoder.dump(self.attrname)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()
		self.attrname = decoder.load()


@register("getslice")
class GetSlice(AST):
	"""
	AST node for getting a slice from a list or string object.

	The object is loaded from the AST node :var:`obj` and the start and stop
	indices from the AST node :var:`index1` and :var:`index2`. :var:`index1`
	and :var:`index2` may also be :const:`None` (for missing slice indices,
	which default to the 0 for the start index and the length of the sequence
	for the end index).
	"""

	fields = AST.fields.union({"obj", "index1", "index2"})

	def __init__(self, location=None, start=None, end=None, obj=None, index1=None, index2=None):
		super().__init__(location, start, end)
		self.obj = obj
		self.index1 = index1
		self.index2 = index2

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} obj={0.obj!r} index1={0.index1!r} index2={0.index2!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("obj=")
				p.pretty(self.obj)
				p.breakable()
				p.text("index1=")
				p.pretty(self.index1)
				p.breakable()
				p.text("index2=")
				p.pretty(self.index2)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		obj = (yield from self.obj.eval(vars))
		if self.index1 is not None:
			index1 = (yield from self.index1.eval(vars))
			if self.index2 is not None:
				index2 = (yield from self.index2.eval(vars))
				return obj[index1:index2]
			else:
				return obj[index1:]
		else:
			if self.index2 is not None:
				index2 = (yield from self.index2.eval(vars))
				return obj[:index2]
			else:
				return obj[:]

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)
		encoder.dump(self.index1)
		encoder.dump(self.index2)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()
		self.index1 = decoder.load()
		self.index2 = decoder.load()


class Unary(AST):
	"""
	Base class for all AST nodes implementing unary operators.
	"""

	fields = AST.fields.union({"obj"})

	def __init__(self, location=None, start=None, end=None, obj=None):
		super().__init__(location, start, end)
		self.obj = obj

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.obj!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.pretty(self.obj)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()

	@handleeval
	def eval(self, vars):
		obj = (yield from self.obj.eval(vars))
		return self.evalfold(obj)

	@classmethod
	def make(cls, location, start, end, obj):
		if isinstance(obj, Const):
			result = cls.evalfold(obj.value)
			if not isinstance(result, Undefined):
				return Const(location, start, end, result)
		return cls(location, start, end, obj)


@register("not")
class Not(Unary):
	"""
	AST node for the unary ``not`` operator.
	"""


	@classmethod
	def evalfold(cls, obj):
		return not obj


@register("neg")
class Neg(Unary):
	"""
	AST node for the unary negation (i.e. "-") operator.
	"""


	@classmethod
	def evalfold(cls, obj):
		return -obj


@register("print")
class Print(Unary):
	"""
	AST node for a ``<?print?>`` tag.
	"""

	def _str(self):
		yield "print "
		yield from super()._str()

	@handleeval
	def eval(self, vars):
		yield _str((yield from self.obj.eval(vars)))


@register("printx")
class PrintX(Unary):
	"""
	AST node for a ``<?printx?>`` tag.
	"""

	def _str(self):
		yield "printx "
		yield from super()._str()

	@handleeval
	def eval(self, vars):
		yield _xmlescape((yield from self.obj.eval(vars)))


@register("return")
class Return(Unary):
	"""
	AST node for a ``<?return?>`` tag.
	"""

	def _str(self):
		yield "return "
		yield from super()._str()

	@handleeval
	def eval(self, vars):
		value = (yield from self.obj.eval(vars))
		raise ReturnException(value)


class Binary(AST):
	"""
	Base class for all AST nodes implementing binary operators.
	"""

	fields = AST.fields.union({"obj1", "obj2"})

	def __init__(self, location=None, start=None, end=None, obj1=None, obj2=None):
		super().__init__(location, start, end)
		self.obj1 = obj1
		self.obj2 = obj2

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.obj1!r} {0.obj2!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.pretty(self.obj1)
				p.breakable()
				p.pretty(self.obj2)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj1)
		encoder.dump(self.obj2)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj1 = decoder.load()
		self.obj2 = decoder.load()

	@handleeval
	def eval(self, vars):
		obj1 = (yield from self.obj1.eval(vars))
		obj2 = (yield from self.obj2.eval(vars))
		return self.evalfold(obj1, obj2)

	@classmethod
	def make(cls, location, start, end, obj1, obj2):
		if isinstance(obj1, Const) and isinstance(obj2, Const):
			result = cls.evalfold(obj1.value, obj2.value)
			if not isinstance(result, Undefined):
				return Const(location, start, end, result)
		return cls(location, start, end, obj1, obj2)


@register("getitem")
class GetItem(Binary):
	"""
	AST node for subscripting operator.

	The object (which must be a list, string or dict) is loaded from the AST
	node :var:`obj1` and the index/key is loaded from the AST node :var:`obj2`.
	"""


	@classmethod
	def evaluate(cls, obj1, obj2):
		return obj1[obj2]

	@classmethod
	def evalfold(cls, obj1, obj2):
		try:
			return obj1[obj2]
		except KeyError:
			return UndefinedKey(obj2)
		except IndexError:
			return UndefinedIndex(obj2)


@register("eq")
class EQ(Binary):
	"""
	AST node for the binary ``==`` comparison operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 == obj2


@register("ne")
class NE(Binary):
	"""
	AST node for the binary ``!=`` comparison operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 != obj2


@register("lt")
class LT(Binary):
	"""
	AST node for the binary ``<`` comparison operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 < obj2


@register("le")
class LE(Binary):
	"""
	AST node for the binary ``<=`` comparison operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 <= obj2


@register("gt")
class GT(Binary):
	"""
	AST node for the binary ``>`` comparison operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 > obj2


@register("ge")
class GE(Binary):
	"""
	AST node for the binary ``>=`` comparison operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 >= obj2


@register("contains")
class Contains(Binary):
	"""
	AST node for the binary containment testing operator.

	The item/key object is loaded from the AST node :var:`obj1` and the container
	object (which must be a list, string or dict) is loaded from the AST node
	:var:`obj2`.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 in obj2


@register("notcontains")
class NotContains(Binary):
	"""
	AST node for the inverted containment testing operator.

	The item/key object is loaded from the AST node :var:`obj1` and the container
	object (which must be a list, string or dict) is loaded from the AST node
	:var:`obj2`.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 not in obj2


@register("add")
class Add(Binary):
	"""
	AST node for the binary addition operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 + obj2


@register("sub")
class Sub(Binary):
	"""
	AST node for the binary substraction operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 - obj2


@register("mul")
class Mul(Binary):
	"""
	AST node for the binary multiplication operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 * obj2


@register("floordiv")
class FloorDiv(Binary):
	"""
	AST node for the binary truncating division operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 // obj2


@register("truediv")
class TrueDiv(Binary):
	"""
	AST node for the binary true division operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 / obj2


@register("and")
class And(Binary):
	"""
	AST node for the binary ``and`` operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		# This is not called from ``eval``, as it doesn't short-circuit
		return obj1 and obj2

	@handleeval
	def eval(self, vars):
		obj1 = (yield from self.obj1.eval(vars))
		if not obj1:
			return obj1
		return (yield from self.obj2.eval(vars))


@register("or")
class Or(Binary):
	"""
	AST node for the binary ``or`` operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		# This is not called from ``eval``, as it doesn't short-circuit
		return obj1 or obj2

	@handleeval
	def eval(self, vars):
		obj1 = (yield from self.obj1.eval(vars))
		if obj1:
			return obj1
		return (yield from self.obj2.eval(vars))


@register("mod")
class Mod(Binary):
	"""
	AST node for the binary modulo operator.
	"""


	@classmethod
	def evalfold(cls, obj1, obj2):
		return obj1 % obj2


class ChangeVar(AST):
	"""
	Baseclass for all AST nodes that store or modify a variable.

	The variable name is stored in the string :var:`varname` and the value that
	will be stored or be used to modify the stored value is loaded from the
	AST node :var:`value`.
	"""

	fields = AST.fields.union({"varname", "value"})

	def __init__(self, location=None, start=None, end=None, varname=None, value=None):
		super().__init__(location, start, end)
		self.varname = varname
		self.value = value

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} varname={0.varname!r} value={0.value!r} at {1:#x}>".format(self, id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("varname=")
				p.pretty(self.varname)
				p.breakable()
				p.text("value=")
				p.pretty(self.value)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.varname)
		encoder.dump(self.value)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.varname = decoder.load()
		self.value = decoder.load()


@register("storevar")
class StoreVar(ChangeVar):
	"""
	AST node that stores a value into a variable.
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		_unpackvar(vars, self.varname, value)


@register("addvar")
class AddVar(ChangeVar):
	"""
	AST node that adds a value to a variable (i.e. the ``+=`` operator).
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		vars[self.varname] += value


@register("subvar")
class SubVar(ChangeVar):
	"""
	AST node that substracts a value from a variable (i.e. the ``-=`` operator).
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		vars[self.varname] -= value


@register("mulvar")
class MulVar(ChangeVar):
	"""
	AST node that multiplies a variable by a value (i.e. the ``*=`` operator).
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		vars[self.varname] *= value


@register("floordivvar")
class FloorDivVar(ChangeVar):
	"""
	AST node that divides a variable by a value (truncating to an integer value;
	i.e. the ``//=`` operator).
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		vars[self.varname] //= value


@register("truedivvar")
class TrueDivVar(ChangeVar):
	"""
	AST node that divides a variable by a value (i.e. the ``/=`` operator).
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		vars[self.varname] /= value


@register("modvar")
class ModVar(ChangeVar):
	"""
	AST node for the ``%=`` operator.
	"""

	@handleeval
	def eval(self, vars):
		value = (yield from self.value.eval(vars))
		vars[self.varname] %= value


@register("callfunc")
class CallFunc(AST):
	"""
	AST node for calling an function.

	The object to be called is stored in the attribute :var:`obj`. The list of
	positional arguments is loaded from the list of AST nodes :var:`args`.
	Keyword arguments are in :var:`kwargs`. `var`:remargs` is the AST node
	for the ``*`` argument (and may by ``None`` if there is no ``*`` argument).
	`var`:remkwargs` is the AST node for the ``**`` argument (and may by ``None``
	if there is no ``**`` argument)
	"""

	fields = AST.fields.union({"obj", "args", "kwargs", "remargs", "remkwargs"})

	def __init__(self, location=None, start=None, end=None, obj=None):
		super().__init__(location, start, end)
		self.obj = obj
		self.args = []
		self.kwargs = []
		self.remargs = None
		self.remkwargs = None

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} obj={0.obj!r}{1}{2}{3}{4} at {5:#x}>".format(
			self,
			"".join(" {!r}".format(arg) for arg in self.args),
			"".join(" {}={!r}".format(argname, argvalue) for (argname, argvalue) in self.kwargs),
			" *{!r}".format(self.remargs) if self.remargs is not None else "",
			" **{!r}".format(self.remkwargs) if self.remargs is not None else "",
			id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("obj=")
				p.pretty(self.obj)
				for arg in self.args:
					p.breakable()
					p.pretty(arg)
				for (argname, arg) in self.kwargs:
					p.breakable()
					p.text("{}=".format(argname))
					p.pretty(arg)
				if self.remargs is not None:
					p.breakable()
					p.text("*")
					p.pretty(self.remargs)
				if self.remkwargs is not None:
					p.breakable()
					p.text("**")
					p.pretty(self.remkwargs)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		obj = (yield from self.obj.eval(vars))
		args = []
		for arg in self.args:
			arg = (yield from arg.eval(vars))
			args.append(arg)
		kwargs = {}
		for (argname, arg) in self.kwargs:
			kwargs[argname] = (yield from arg.eval(vars))
		if self.remargs is not None:
			args.extend((yield from self.remargs.eval(vars)))
		if self.remkwargs is not None:
			kwargs.update((yield from self.remkwargs.eval(vars)))
		result = obj(*args, **kwargs)
		if isinstance(obj, (Template, TemplateClosure)):
			return result
		else:
			return (yield from result)

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.obj)
		encoder.dump(self.args)
		encoder.dump(self.kwargs)
		encoder.dump(self.remargs)
		encoder.dump(self.remkwargs)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.obj = decoder.load()
		self.args = decoder.load()
		self.kwargs = [tuple(arg) for arg in decoder.load()]
		self.remargs = decoder.load()
		self.remkwargs = decoder.load()


@register("callmeth")
class CallMeth(AST):
	"""
	AST node for calling a method.

	The method name is stored in the string :var:`methname`. The object for which
	the method will be called is loaded from the AST node :var:`obj` and the list
	of arguments is loaded from the list of AST nodes :var:`args`. Keyword
	arguments are in :var:`kwargs`. `var`:remargs` is the AST node for the ``*``
	argument (and may by ``None`` if there is no ``*`` argument).
	`var`:remkwargs` is the AST node for the ``**`` argument (and may by ``None``
	if there is no ``**`` argument)
	"""

	fields = AST.fields.union({"obj", "methname", "args", "kwargs", "remargs", "remkwargs"})

	def __init__(self, location=None, start=None, end=None, obj=None, methname=None):
		super().__init__(location, start, end)
		self.obj = obj
		self.methname = methname
		self.args = []
		self.kwargs = []
		self.remargs = None
		self.remkwargs = None

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} methname={0.methname!r} obj={0.obj!r}{1}{2}{3}{4} at {5:#x}>".format(
			self,
			"".join(" {!r}".format(arg) for arg in self.args),
			"".join(" {}={!r}".format(argname, argvalue) for (argname, argvalue) in self.kwargs),
			" *{!r}".format(self.remargs) if self.remargs is not None else "",
			" **{!r}".format(self.remkwargs) if self.remargs is not None else "",
			id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("methname=")
				p.pretty(self.methname)
				p.breakable()
				p.text("obj=")
				p.pretty(self.obj)
				for arg in self.args:
					p.breakable()
					p.pretty(arg)
				for (argname, arg) in self.kwargs:
					p.breakable()
					p.text("{}=".format(argname))
					p.pretty(arg)
				if self.remargs is not None:
					p.breakable()
					p.text("*")
					p.pretty(self.remargs)
				if self.remkwargs is not None:
					p.breakable()
					p.text("**")
					p.pretty(self.remkwargs)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	@handleeval
	def eval(self, vars):
		obj = (yield from self.obj.eval(vars))
		args = []
		for arg in self.args:
			arg = (yield from arg.eval(vars))
			args.append(arg)
		kwargs = {}
		for (argname, arg) in self.kwargs:
			kwargs[argname] = (yield from arg.eval(vars))
		if self.remargs is not None:
			args.extend((yield from self.remargs.eval(vars)))
		if self.remkwargs is not None:
			kwargs.update((yield from self.remkwargs.eval(vars)))
		result = self.methods[self.methname](obj, *args, **kwargs)
		if isinstance(result, types.GeneratorType):
			return (yield from result)
		else:
			return result

	def ul4ondump(self, encoder):
		super().ul4ondump(encoder)
		encoder.dump(self.methname)
		encoder.dump(self.obj)
		encoder.dump(self.args)
		encoder.dump(self.kwargs)
		encoder.dump(self.remargs)
		encoder.dump(self.remkwargs)

	def ul4onload(self, decoder):
		super().ul4onload(decoder)
		self.methname = decoder.load()
		self.obj = decoder.load()
		self.args = decoder.load()
		self.kwargs = [tuple(arg) for arg in decoder.load()]
		self.remargs = decoder.load()
		self.remkwargs = decoder.load()


@register("template")
class Template(Block):
	"""
	A template object is normally created by passing the template source to the
	constructor. It can also be loaded from the compiled format via the class
	methods :meth:`load` (from a stream) or :meth:`loads` (from a string).

	The compiled format can be generated with the methods :meth:`dump` (which
	dumps the format to a stream) or :meth:`dumps` (which returns a string with
	the compiled format).

	Rendering the template can be done with the methods :meth:`render` (which
	is a generator) or :meth:`renders` (which returns a string).

	A :class:`Template` object is itself an AST node. Evaluating it will store
	the template object under its name in the local variables.

	A :class:`Template` can also be called as a function (returning the result
	of the first ``<?return?>`` tag encountered. In this case all output of the
	template will be ignored.
	"""
	fields = Block.fields.union({"source", "name", "keepws", "startdelim", "enddelim"})

	version = "24"

	def __init__(self, source=None, name=None, keepws=True, startdelim="<?", enddelim="?>"):
		"""
		Create a :class:`Template` object. If :var:`source` is ``None``, the
		:class:`Template` remains uninitialized, otherwise :var:`source` will be
		compiled (using :var:`startdelim` and :var:`enddelim` as the tag
		delimiters). :var:`name` is the name of the template. It will be used in
		exception messages and should be a valid Python identifier. If
		:var:`keepws` is false linefeeds and indentation will be ignored in the
		literal text in templates (i.e. the text between the tags). However
		trailing whitespace at the end of the line will be honored regardless of
		the value of :var:`keepws`. Output will always be ignored when calling
		a template as a function.
		"""
		# ``location``/``endlocation`` will remain ``None`` for a top level template
		# For a subtemplate/subfunction ``location`` will be set to the location of the ``<?def?>`` tag in :meth:`_compile`
		# and ``endlocation`` will be the location of the ``<?end def?>`` tag
		super().__init__(None, 0, 0)
		self.keepws = keepws
		self.startdelim = startdelim
		self.enddelim = enddelim
		self.name = name
		self.source = None

		# If we have source code compile it
		if source is not None:
			self._compile(source, name, startdelim, enddelim)

	def __repr__(self):
		s = "<{0.__class__.__module__}.{0.__class__.__qualname__} name={0.name!r} keepws={0.keepws!r}".format(self)
		if self.startdelim != "<?":
			s += " startdelim={0.startdelim!r}".format(self)
		if self.enddelim != "?>":
			s += " enddelim={0.enddelim!r}".format(self)
		if self.content:
			s + " ..."
		return s + " at {:#x}>".format(id(self))

	def _str(self):
		yield "def "
		yield self.name if self.name is not None else "unnamed"
		yield ":"
		yield None
		yield +1
		yield from super()._str()
		yield -1

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("name=")
				p.pretty(self.name)
				p.breakable()
				p.text("keepws=")
				p.pretty(self.keepws)
				if self.startdelim != "<?":
					p.breakable()
					p.text("startdelim=")
					p.pretty(self.startdelim)
				if self.enddelim != "?>":
					p.breakable()
					p.text("enddelim=")
					p.pretty(self.enddelim)
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))

	def ul4ondump(self, encoder):
		# Don't call ``super().ul4ondump()`` first, as we want the version to be first
		encoder.dump(self.version)
		encoder.dump(self.source)
		encoder.dump(self.name)
		encoder.dump(self.keepws)
		encoder.dump(self.startdelim)
		encoder.dump(self.enddelim)
		super().ul4ondump(encoder)

	def ul4onload(self, decoder):
		version = decoder.load()
		if version != self.version:
			raise ValueError("invalid version, expected {!r}, got {!r}".format(self.version, version))
		self.source = decoder.load()
		self.name = decoder.load()
		self.keepws = decoder.load()
		self.startdelim = decoder.load()
		self.enddelim = decoder.load()
		super().ul4onload(decoder)

	@classmethod
	def loads(cls, data):
		"""
		The class method :meth:`loads` loads the template/function from string
		:var:`data`. :var:`data` must contain the template/function in compiled
		UL4ON format.
		"""
		from ll import ul4on
		return ul4on.loads(data)

	@classmethod
	def load(cls, stream):
		"""
		The class method :meth:`load` loads the template/function from the stream
		:var:`stream`. The stream must contain the template/function in compiled
		UL4ON format.
		"""
		from ll import ul4on
		return ul4on.load(stream)

	def dump(self, stream):
		"""
		:meth:`dump` dumps the template/function in compiled UL4ON format to the
		stream :var:`stream`.
		"""
		from ll import ul4on
		ul4on.dump(self, stream)

	def dumps(self):
		"""
		:meth:`dumps` returns the template/function in compiled UL4ON format
		(as a string).
		"""
		from ll import ul4on
		return ul4on.dumps(self)

	def render(self, **vars):
		"""
		Render the template iteratively (i.e. this is a generator).
		:var:`vars` contains the top level variables available to the
		template code.
		"""
		try:
			yield from super().eval(vars) # Bypass ``self.eval()`` which simply stores the object as a local variable
		except ReturnException:
			pass

	def renders(self, **vars):
		"""
		Render the template as a string. :var:`vars` contains the top level
		variables available to the template code.
		"""
		return "".join(self.render(**vars))

	def __call__(self, **vars):
		"""
		Call the template as a function and return the resulting value.
		:var:`vars` contains the top level variables available to the template code.
		"""
		try:
			for output in super().eval(vars): # Bypass ``self.eval()`` which simply stores the object as a local variable
				pass # Ignore all output
		except ReturnException as ex:
			return ex.value

	def jssource(self):
		"""
		Return the template as the source code of a Javascript function.
		"""
		return "ul4.Template.loads({})".format(_asjson(self.dumps()))

	def javasource(self):
		"""
		Return the template as Java source code.
		"""
		return "com.livinglogic.ul4.InterpretedTemplate.loads({})".format(misc.javaexpr(self.dumps()))

	def _tokenize(self, source, startdelim, enddelim):
		"""
		Tokenize the template/function source code :var:`source` into tags and
		non-tag text. :var:`startdelim` and :var:`enddelim` are used as the tag
		delimiters.

		This is a generator which produces :class:`Location` objects for each tag
		or non-tag text. It will be called by :meth:`_compile` internally.
		"""
		pattern = "{}(printx|print|code|for|if|elif|else|end|break|continue|def|return|note)(\s*((.|\\n)*?)\s*)?{}".format(re.escape(startdelim), re.escape(enddelim))
		pos = 0
		for match in re.finditer(pattern, source):
			if match.start() != pos:
				yield Location(self, source, None, pos, match.start(), pos, match.start())
			type = source[match.start(1):match.end(1)]
			if type != "note":
				yield Location(self, source, type, match.start(), match.end(), match.start(3), match.end(3))
			pos = match.end()
		end = len(source)
		if pos != end:
			yield Location(self, source, None, pos, end, pos, end)

	def _parser(self, location, error):
		from ll import UL4Lexer, UL4Parser
		source = location.code
		if not source:
			raise ValueError(error)
		stream = antlr3.ANTLRStringStream(source)
		lexer = UL4Lexer.UL4Lexer(stream)
		lexer.location = location
		tokens = antlr3.CommonTokenStream(lexer)
		parser = UL4Parser.UL4Parser(tokens)
		parser.location = location
		return parser

	def _compile(self, source, name, startdelim, enddelim):
		"""
		Compile the template source code :var:`source` into an AST.
		:var:`startdelim` and :var:`enddelim` are used as the tag delimiters.
		"""
		self.name = name
		self.startdelim = startdelim
		self.enddelim = enddelim

		# This stack stores the nested for/if/elif/else/def blocks
		stack = [self]

		self.source = source

		if source is None:
			return

		def parseexpr(location):
			return self._parser(location, "expression required").expression()

		def parsestmt(location):
			return self._parser(location, "statement required").statement()

		def parsefor(location):
			return self._parser(location, "loop expression required").for_()

		for location in self._tokenize(source, startdelim, enddelim):
			try:
				if location.type is None:
					stack[-1].append(Text(location, location.startcode, location.endcode))
				elif location.type == "print":
					stack[-1].append(Print(location, location.startcode, location.endcode, parseexpr(location)))
				elif location.type == "printx":
					stack[-1].append(PrintX(location, location.startcode, location.endcode, parseexpr(location)))
				elif location.type == "code":
					stack[-1].append(parsestmt(location))
				elif location.type == "if":
					block = IfElIfElse(location, location.startcode, location.endcode, parseexpr(location))
					stack[-1].append(block)
					stack.append(block)
				elif location.type == "elif":
					if not isinstance(stack[-1], IfElIfElse):
						raise BlockError("elif doesn't match and if")
					elif isinstance(stack[-1].content[-1], Else):
						raise BlockError("else already seen in if")
					stack[-1].newblock(ElIf(location, location.startcode, location.endcode, parseexpr(location)))
				elif location.type == "else":
					if not isinstance(stack[-1], IfElIfElse):
						raise BlockError("else doesn't match any if")
					elif isinstance(stack[-1].content[-1], Else):
						raise BlockError("else already seen in if")
					stack[-1].newblock(Else(location, location.startcode, location.endcode))
				elif location.type == "end":
					if len(stack) <= 1:
						raise BlockError("not in any block")
					code = location.code
					if code:
						if code == "if":
							if not isinstance(stack[-1], IfElIfElse):
								raise BlockError("endif doesn't match any if")
						elif code == "for":
							if not isinstance(stack[-1], For):
								raise BlockError("endfor doesn't match any for")
						elif code == "def":
							if not isinstance(stack[-1], Template):
								raise BlockError("enddef doesn't match any def")
						else:
							raise BlockError("illegal end value {!r}".format(code))
					last = stack.pop()
					# Set ``endlocation`` of block
					last.endlocation = location
					if isinstance(last, IfElIfElse):
						last.content[-1].endlocation = location
				elif location.type == "for":
					block = parsefor(location)
					stack[-1].append(block)
					stack.append(block)
				elif location.type == "break":
					for block in reversed(stack):
						if isinstance(block, For):
							break
						elif isinstance(block, Template):
							raise BlockError("break outside of for loop")
					stack[-1].append(Break(location, location.startcode, location.endcode))
				elif location.type == "continue":
					for block in reversed(stack):
						if isinstance(block, For):
							break
						elif isinstance(block, Template):
							raise BlockError("continue outside of for loop")
					stack[-1].append(Continue(location, location.startcode, location.endcode))
				elif location.type == "def":
					block = Template(None, location.code, self.keepws, self.startdelim, self.enddelim)
					block.location = location # Set start ``location`` of sub template
					block.source = self.source # The source of the top level template (so that the offsets in :class:`Location` are correct)
					block.start = location.startcode
					block.end = location.endcode
					stack[-1].append(block)
					stack.append(block)
				elif location.type == "return":
					stack[-1].append(Return(location, location.startcode, location.endcode, parseexpr(location)))
				else: # Can't happen
					raise ValueError("unknown tag {!r}".format(location.type))
			except Exception as exc:
				raise Error(self) from exc
		if len(stack) > 1:
			raise Error(stack[-1]) from BlockError("block unclosed")

	@handleeval
	def eval(self, vars):
		yield from ()
		vars[self.name] = TemplateClosure(self, vars)


###
### Functions & methods
###

@AST.makefunction
def function_print(*values):
	for (i, value) in enumerate(values):
		if i:
			yield " "
		yield _str(value)


@AST.makefunction
def function_printx(*values):
	for (i, value) in enumerate(values):
		if i:
			yield " "
		yield _xmlescape(value)


@AST.makefunction
def function_str(obj=""):
	yield from ()
	return _str(obj)


@AST.makefunction
def function_repr(obj):
	yield from ()
	return _repr(obj)


@AST.makefunction
def function_now():
	yield from ()
	return datetime.datetime.now()


@AST.makefunction
def function_utcnow():
	yield from ()
	return datetime.datetime.utcnow()


@AST.makefunction
def function_date(year, month, day, hour=0, minute=0, second=0, microsecond=0):
	yield from ()
	return datetime.datetime(year, month, day, hour, minute, second, microsecond)


@AST.makefunction
def function_timedelta(days=0, seconds=0, microseconds=0):
	yield from ()
	return datetime.timedelta(days, seconds, microseconds)


@AST.makefunction
def function_monthdelta(months=0):
	yield from ()
	return misc.monthdelta(months)


@AST.makefunction
def function_random():
	yield from ()
	return random.random()


@AST.makefunction
def function_xmlescape(obj):
	yield from ()
	return _xmlescape(obj)


@AST.makefunction
def function_csv(obj):
	yield from ()
	if obj is None:
		return ""
	elif isinstance(obj, Undefined):
		return ""
	elif not isinstance(obj, str):
		obj = _repr(obj)
	if any(c in obj for c in ',"\n'):
		return '"{}"'.format(obj.replace('"', '""'))
	return obj


@AST.makefunction
def function_asjson(obj):
	yield from ()
	return _asjson(obj)


@AST.makefunction
def function_fromjson(string):
	from ll import ul4on
	yield from ()
	return json.loads(string)


@AST.makefunction
def function_asul4on(obj):
	from ll import ul4on
	yield from ()
	return ul4on.dumps(obj)


@AST.makefunction
def function_fromul4on(string):
	from ll import ul4on
	yield from ()
	return ul4on.loads(string)


@AST.makefunction
def function_int(obj=0, base=None):
	yield from ()
	if base is None:
		return int(obj)
	else:
		return int(obj, base)


@AST.makefunction
def function_float(obj=0.0):
	yield from ()
	return float(obj)


@AST.makefunction
def function_bool(obj=False):
	yield from ()
	return bool(obj)


@AST.makefunction
def function_len(sequence):
	yield from ()
	return len(sequence)


@AST.makefunction
def function_abs(number):
	yield from ()
	return abs(number)


@AST.makefunction
def function_any(iterable):
	yield from ()
	return any(iterable)


@AST.makefunction
def function_all(iterable):
	yield from ()
	return all(iterable)


@AST.makefunction
def function_enumerate(iterable, start=0):
	yield from ()
	return enumerate(iterable, start)


@AST.makefunction
def function_enumfl(iterable, start=0):
	yield from ()
	def result(iterable):
		lastitem = None
		first = True
		i = start
		it = iter(iterable)
		try:
			item = next(it)
		except StopIteration:
			return
		while True:
			try:
				(lastitem, item) = (item, next(it))
			except StopIteration:
				yield (i, first, True, item) # Items haven't been swapped yet
				return
			else:
				yield (i, first, False, lastitem)
				first = False
			i += 1
	return result(iterable)


@AST.makefunction
def function_isfirstlast(iterable):
	yield from ()
	def result(iterable):
		lastitem = None
		first = True
		it = iter(iterable)
		try:
			item = next(it)
		except StopIteration:
			return
		while True:
			try:
				(lastitem, item) = (item, next(it))
			except StopIteration:
				yield (first, True, item) # Items haven't been swapped yet
				return
			else:
				yield (first, False, lastitem)
				first = False
	return result(iterable)


@AST.makefunction
def function_isfirst(iterable):
	yield from ()
	def result(iterable):
		first = True
		for item in iterable:
			yield (first, item)
			first = False
	return result(iterable)


@AST.makefunction
def function_islast(iterable):
	yield from ()
	def result(iterable):
		lastitem = None
		it = iter(iterable)
		try:
			item = next(it)
		except StopIteration:
			return
		while True:
			try:
				(lastitem, item) = (item, next(it))
			except StopIteration:
				yield (True, item) # Items haven't been swapped yet
				return
			else:
				yield (False, lastitem)
	return result(iterable)


@AST.makefunction
def function_isundefined(obj):
	yield from ()
	return isinstance(obj, Undefined)


@AST.makefunction
def function_isdefined(obj):
	yield from ()
	return not isinstance(obj, Undefined)


@AST.makefunction
def function_isnone(obj):
	yield from ()
	return obj is None


@AST.makefunction
def function_isstr(obj):
	yield from ()
	return isinstance(obj, str)


@AST.makefunction
def function_isint(obj):
	yield from ()
	return isinstance(obj, int) and not isinstance(obj, bool)


@AST.makefunction
def function_isfloat(obj):
	yield from ()
	return isinstance(obj, float)


@AST.makefunction
def function_isbool(obj):
	yield from ()
	return isinstance(obj, bool)


@AST.makefunction
def function_isdate(obj):
	yield from ()
	return isinstance(obj, (datetime.datetime, datetime.date))


@AST.makefunction
def function_istimedelta(obj):
	yield from ()
	return isinstance(obj, datetime.timedelta)


@AST.makefunction
def function_ismonthdelta(obj):
	yield from ()
	return isinstance(obj, misc.monthdelta)


@AST.makefunction
def function_islist(obj):
	yield from ()
	return isinstance(obj, collections.Sequence) and not isinstance(obj, str) and not isinstance(obj, color.Color)


@AST.makefunction
def function_isdict(obj):
	yield from ()
	return isinstance(obj, collections.Mapping) and not isinstance(obj, Template)


@AST.makefunction
def function_iscolor(obj):
	yield from ()
	return isinstance(obj, color.Color)


@AST.makefunction
def function_istemplate(obj):
	yield from ()
	return isinstance(obj, (Template, TemplateClosure))


@AST.makefunction
def function_isfunction(obj):
	yield from ()
	return callable(obj)


@AST.makefunction
def function_chr(i):
	yield from ()
	return chr(i)


@AST.makefunction
def function_ord(c):
	yield from ()
	return ord(c)


@AST.makefunction
def function_hex(number):
	yield from ()
	return hex(number)


@AST.makefunction
def function_oct(number):
	yield from ()
	return oct(number)


@AST.makefunction
def function_bin(number):
	yield from ()
	return bin(number)


@AST.makefunction
def function_min(*args):
	yield from ()
	return min(*args)


@AST.makefunction
def function_max(*args):
	yield from ()
	return max(*args)


@AST.makefunction
def function_sorted(iterable):
	yield from ()
	return sorted(iterable)


@AST.makefunction
def function_range(*args):
	yield from ()
	return range(*args)


@AST.makefunction
def function_type(obj):
	yield from ()
	if obj is None:
		return "none"
	elif isinstance(obj, Undefined):
		return "undefined"
	elif isinstance(obj, str):
		return "str"
	elif isinstance(obj, bool):
		return "bool"
	elif isinstance(obj, int):
		return "int"
	elif isinstance(obj, float):
		return "float"
	elif isinstance(obj, (datetime.datetime, datetime.date)):
		return "date"
	elif isinstance(obj, datetime.timedelta):
		return "timedelta"
	elif isinstance(obj, misc.monthdelta):
		return "monthdelta"
	elif isinstance(obj, color.Color):
		return "color"
	elif isinstance(obj, (Template, TemplateClosure)):
		return "template"
	elif isinstance(obj, collections.Mapping):
		return "dict"
	elif isinstance(obj, color.Color):
		return "color"
	elif isinstance(obj, collections.Sequence):
		return "list"
	elif callable(obj):
		return "function"
	return None


@AST.makefunction
def function_reversed(sequence):
	yield from ()
	return reversed(sequence)


@AST.makefunction
def function_randrange(*args):
	yield from ()
	return random.randrange(*args)


@AST.makefunction
def function_randchoice(sequence):
	yield from ()
	return random.choice(sequence)


@AST.makefunction
def function_format(obj, fmt, lang=None):
	yield from ()
	if isinstance(obj, (datetime.date, datetime.time, datetime.timedelta)):
		if lang is None:
			lang = "en"
		oldlocale = locale.getlocale()
		try:
			for candidate in (locale.normalize(lang), locale.normalize("en"), ""):
				try:
					locale.setlocale(locale.LC_ALL, candidate)
					return format(obj, fmt)
				except locale.Error:
					if not candidate:
						return format(obj, fmt)
		finally:
			try:
				locale.setlocale(locale.LC_ALL, oldlocale)
			except locale.Error:
				pass
	else:
		return format(obj, fmt)


@AST.makefunction
def function_zip(*iterables):
	yield from ()
	return zip(*iterables)


@AST.makefunction
def function_urlquote(string):
	yield from ()
	return urlparse.quote_plus(string)


@AST.makefunction
def function_urlunquote(string):
	yield from ()
	return urlparse.unquote_plus(string)


@AST.makefunction
def function_rgb(r, g, b, a=1.0):
	yield from ()
	return color.Color.fromrgb(r, g, b, a)


@AST.makefunction
def function_hls(h, l, s, a=1.0):
	yield from ()
	return color.Color.fromhls(h, l, s, a)


@AST.makefunction
def function_hsv(h, s, v, a=1.0):
	yield from ()
	return color.Color.fromhsv(h, s, v, a)


@AST.makemethod
def method_split(obj, sep=None, count=None):
	yield from ()
	return obj.split(sep, count if count is not None else -1)


@AST.makemethod
def method_rsplit(obj, sep=None, count=None):
	yield from ()
	return obj.rsplit(sep, count if count is not None else -1)


@AST.makemethod
def method_strip(obj, chars=None):
	yield from ()
	return obj.strip(chars)


@AST.makemethod
def method_lstrip(obj, chars=None):
	yield from ()
	return obj.lstrip(chars)


@AST.makemethod
def method_rstrip(obj, chars=None):
	yield from ()
	return obj.rstrip(chars)


@AST.makemethod
def method_find(obj, sub, start=None, end=None):
	yield from ()
	if isinstance(obj, str):
		return obj.find(sub, start, end)
	else:
		try:
			if end is None:
				if start is None:
					return obj.index(sub)
				return obj.index(sub, start)
			return obj.index(sub, start, end)
		except ValueError:
			return -1


@AST.makemethod
def method_rfind(obj, sub, start=None, end=None):
	yield from ()
	if isinstance(obj, str):
		return obj.rfind(sub, start, end)
	else:
		for i in reversed(range(*slice(start, end).indices(len(obj)))):
			if obj[i] == sub:
				return i
		return -1


@AST.makemethod
def method_startswith(obj, prefix):
	yield from ()
	return obj.startswith(prefix)


@AST.makemethod
def method_endswith(obj, suffix):
	yield from ()
	return obj.endswith(suffix)


@AST.makemethod
def method_upper(obj):
	yield from ()
	return obj.upper()


@AST.makemethod
def method_lower(obj):
	yield from ()
	return obj.lower()


@AST.makemethod
def method_capitalize(obj):
	yield from ()
	return obj.capitalize()


@AST.makemethod
def method_replace(obj, old, new, count=None):
	yield from ()
	if count is None:
		return obj.replace(old, new)
	else:
		return obj.replace(old, new, count)


@AST.makemethod
def method_r(obj):
	yield from ()
	return obj.r()


@AST.makemethod
def method_g(obj):
	yield from ()
	return obj.g()


@AST.makemethod
def method_b(obj):
	yield from ()
	return obj.b()


@AST.makemethod
def method_a(obj):
	yield from ()
	return obj.a()


@AST.makemethod
def method_hls(obj):
	yield from ()
	return obj.hls()


@AST.makemethod
def method_hlsa(obj):
	yield from ()
	return obj.hlsa()


@AST.makemethod
def method_hsv(obj):
	yield from ()
	return obj.hsv()


@AST.makemethod
def method_hsva(obj):
	yield from ()
	return obj.hsva()


@AST.makemethod
def method_lum(obj):
	yield from ()
	return obj.lum()


@AST.makemethod
def method_weekday(obj):
	yield from ()
	return obj.weekday()


@AST.makemethod
def method_week(obj, firstweekday=None):
	yield from ()
	if firstweekday is None:
		firstweekday = 0
	else:
		firstweekday %= 7
	jan1 = obj.__class__(obj.year, 1, 1)
	yearday = (obj - jan1).days+7
	jan1weekday = jan1.weekday()
	while jan1weekday != firstweekday:
		yearday -= 1
		jan1weekday += 1
		if jan1weekday == 7:
			jan1weekday = 0
	return yearday//7


@AST.makemethod
def method_items(obj):
	yield from ()
	return obj.items()


@AST.makemethod
def method_values(obj):
	yield from ()
	return obj.values()


@AST.makemethod
def method_join(obj, iterable):
	yield from ()
	return obj.join(iterable)


@AST.makemethod
def method_render(obj, **vars):
	yield from obj.render(**vars)


@AST.makemethod
def method_renders(obj, **vars):
	yield from ()
	return obj.renders(**vars)


@AST.makemethod
def method_mimeformat(obj):
	yield from ()
	weekdayname = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
	monthname = (None, "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")
	return "{1}, {0.day:02d} {2:3} {0.year:4} {0.hour:02}:{0.minute:02}:{0.second:02} GMT".format(obj, weekdayname[obj.weekday()], monthname[obj.month])


@AST.makemethod
def method_isoformat(obj):
	yield from ()
	result = obj.isoformat()
	suffix = "T00:00:00"
	if result.endswith(suffix):
		return result[:-len(suffix)]
	return result


@AST.makemethod
def method_yearday(obj):
	yield from ()
	return (obj - obj.__class__(obj.year, 1, 1)).days+1


@AST.makemethod
def method_get(obj, key, default=None):
	yield from ()
	return obj.get(key, default)


@AST.makemethod
def method_withlum(obj, lum):
	yield from ()
	return obj.withlum(lum)


@AST.makemethod
def method_witha(obj, a):
	yield from ()
	return obj.witha(a)


@AST.makemethod
def method_day(obj):
	yield from ()
	return obj.day


@AST.makemethod
def method_month(obj):
	return obj.month


@AST.makemethod
def method_year(obj):
	yield from ()
	return obj.year


@AST.makemethod
def method_hour(obj):
	yield from ()
	return obj.hour


@AST.makemethod
def method_minute(obj):
	yield from ()
	return obj.minute


@AST.makemethod
def method_second(obj):
	yield from ()
	return obj.second


@AST.makemethod
def method_microsecond(obj):
	yield from ()
	return obj.microsecond


@AST.makemethod
def method_days(obj):
	yield from ()
	return obj.days


@AST.makemethod
def method_seconds(obj):
	yield from ()
	return obj.seconds


@AST.makemethod
def method_microseconds(obj):
	yield from ()
	return obj.microseconds


@AST.makemethod
def method_months(obj):
	yield from ()
	return obj.months


@AST.makemethod
def method_append(obj, *items):
	yield from ()
	obj.extend(items)


@AST.makemethod
def method_insert(obj, pos, *items):
	yield from ()
	obj[pos:pos] = items


@AST.makemethod
def method_pop(obj, pos=-1):
	yield from ()
	return obj.pop(pos)


@AST.makemethod
def method_update(obj, *others, **kwargs):
	yield from ()
	for other in others:
		obj.update(other)
	obj.update(**kwargs)


class TemplateClosure(Object):
	fields = {"location", "endlocation", "name", "source", "startdelim", "enddelim", "content"}

	def __init__(self, template, vars):
		self.template = template
		# Freeze variables of the currently running templates/functions
		self.vars = vars.copy()
		# The template (i.e. the closure) itself should be visible in the parent variables
		self.vars[template.name] = self

	def render(self, **vars):
		return self.template.render(**collections.ChainMap(vars, self.vars))

	def renders(self, **vars):
		return self.template.renders(**collections.ChainMap(vars, self.vars))

	def __call__(self, **vars):
		return self.template(**collections.ChainMap(vars, self.vars))

	def __getattr__(self, name):
		return getattr(self.template, name)

	def __repr__(self):
		s = "<{0.__class__.__module__}.{0.__class__.__qualname__} name={0.name!r} keepws={0.keepws!r}".format(self)
		if self.startdelim != "<?":
			s += " startdelim={0.startdelim!r}".format(self)
		if self.enddelim != "?>":
			s += " enddelim={0.enddelim!r}".format(self)
		if self.content:
			s + " ..."
		return s + " at {:#x}>".format(id(self))

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text("<{0.__class__.__module__}.{0.__class__.__qualname__} ... at {1:#x}>".format(self, id(self)))
		else:
			with p.group(4, "<{0.__class__.__module__}.{0.__class__.__qualname__}".format(self), ">"):
				p.breakable()
				p.text("name=")
				p.pretty(self.name)
				p.breakable()
				p.text("keepws=")
				p.pretty(self.keepws)
				if self.startdelim != "<?":
					p.breakable()
					p.text("startdelim=")
					p.pretty(self.startdelim)
				if self.enddelim != "?>":
					p.breakable()
					p.text("enddelim=")
					p.pretty(self.enddelim)
				for node in self.content:
					p.breakable()
					p.pretty(node)
				p.breakable()
				p.text("at {:#x}".format(id(self)))



#! /usr/bin/env/python
# -*- coding: utf-8 -*-

## Copyright 1999-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


import types

from ll.xist import xsc, xnd, sims


xmlns = "http://xmlns.example.com/"


def xnd2ns(data):
	with xsc.Pool(): # don't pollute the defaultpool
		code = str(data)
		code = compile(code, "test.py", "exec")
		mod = types.ModuleType("test")
		mod.__file__ = "test.py"
		exec(code, mod.__dict__)
		return mod


def test_xmlns():
	ns = xnd.Module(xmlns)
	ns = xnd2ns(ns)
	assert ns.xsc is xsc


def test_element():
	ns = xnd.Module()
	ns += xnd.Element("foo", xmlns="http://xmlns.foo.com")
	ns += xnd.Element("foo", xmlns="http://xmlns.foo2.com")
	ns = xnd2ns(ns)
	assert ns.foo.xmlname == "foo"
	assert ns.foo.xmlns == "http://xmlns.foo.com"
	assert ns.foo2.xmlname == "foo"
	assert ns.foo2.xmlns == "http://xmlns.foo2.com"


def test_procinst():
	ns = xnd.Module()
	ns += xnd.ProcInst("foo", doc="gurk")
	ns = xnd2ns(ns)

	assert issubclass(ns.foo, xsc.ProcInst)
	assert ns.foo.__doc__.strip() == "gurk"

	ns = xnd.Module(xmlns)
	ns += xnd.ProcInst("f-o-o")
	ns = xnd2ns(ns)
	assert issubclass(ns.f_o_o, xsc.ProcInst)
	assert ns.f_o_o.xmlname == "f-o-o"


def test_entity():
	ns = xnd.Module()
	ns += xnd.Entity("foo", doc="gurk")
	ns = xnd2ns(ns)
	assert issubclass(ns.foo, xsc.Entity)
	assert ns.foo.__doc__.strip() == "gurk"

	ns = xnd.Module()
	ns += xnd.Entity("f-o-o")
	ns = xnd2ns(ns)
	assert issubclass(ns.f_o_o, xsc.Entity)
	assert ns.f_o_o.xmlname == "f-o-o"


def test_charref():
	ns = xnd.Module()
	ns += xnd.CharRef("foo", doc="gurk", codepoint=0x3042)
	ns = xnd2ns(ns)
	assert issubclass(ns.foo, xsc.CharRef)
	assert ns.foo.__doc__.strip() == "gurk"
	assert ns.foo.codepoint == 0x3042

	ns = xnd.Module()
	ns += xnd.CharRef("f-o-o", codepoint=0x3042)
	ns = xnd2ns(ns)
	assert issubclass(ns.f_o_o, xsc.CharRef)
	assert ns.f_o_o.xmlname == "f-o-o"


def test_model():
	ns = xnd.Module(xmlns)
	ns += xnd.Element("foo", modeltype=True)
	ns = xnd2ns(ns)
	assert isinstance(ns.foo.model, sims.Any)

	ns = xnd.Module()
	ns += xnd.Element("foo", modeltype=False)
	ns = xnd2ns(ns)
	assert isinstance(ns.foo.model, sims.Empty)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from ll.xist import xsc, parse
from ll.xist.ns import xml, html, meta


import qel_xmlns, rdf_xmlns, rdfs_xmlns, cc_xmlns, dc_xmlns

url = "http://www.amk.ca/quotations/python-quotes.xml"


if __name__ == "__main__":
	pool = xsc.Pool(html, xml, qel_xmlns, rdf_xmlns, rdfs_xmlns, cc_xmlns, dc_xmlns)
	base = "root:python-quotes.html"
	e = parse.tree(parse.URL(url), parse.Expat(ns=True), parse.Node(pool=pool, base=base), validate=False)
	e = e[qel_xmlns.quotations][0]
	e = e.compacted().conv()
	print(e.string(base=base, encoding="iso-8859-1", validate=False))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

## Copyright 2012-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 2012-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See LICENSE for the license

"""
:mod:`oradd` can be used to import data into an Oracle database. The data is
imported by calling stored procedures.

Basic usage
-----------

Creating an ``oradd`` file can be done like this::

	from ll import oradd

	with open("data.oradd", "w", encoding="utf-8") as f:
		per_id = oradd.Key()
		oradd.dump_oradd(
			f,
			"person_insert",
			per_id=per_id,
			per_firstname=u"Max",
			per_lastname=u"Mustermann"
		)
		oradd.dump_oradd(
			f,
			"contact_insert",
			con_id=oradd.Key(),
			per_id=per_id,
			con_type=u"email",
			con_value=u"max@example.org"
		)

The content of the generated file ``data.oradd`` will look like this::

	{'keys': ['per_id'], 'args': {'per_id': 0, 'per_firstname': 'Max', 'per_lastname': 'Mustermann'}, 'name': 'person_insert'}
	{'keys': ['per_id', 'con_id'], 'args': {'per_id': 0, 'con_id': 1, 'con_type': 'email', 'con_value': 'max@example.org'}, 'name': 'contact_insert'}

i.e. it's just one Python ``repr`` of a dictionary per line (there's also the
option of using UL4ON as the output format).

This file can then be imported into an Oracle database with the following
command::

	python oradd.py <data.pydd user/pwd@database

This will import two records, one by calling ``person_insert`` and one by
calling ``contact_insert``. The PL/SQL equivalent of the above is::

	declare
		v_per_id integer;
		v_con_id integer;
	begin
		person_insert(
			per_id=v_per_id,
			per_firstname='Max',
			per_lastname='Mustermann'
		);
		contact_insert(
			con_id=v_con_id,
			per_id=v_per_id,
			con_type='email',
			con_value='max@example.org'
		)
	end;


Data format
-----------

An oradd file (in its native format) contains one line for each procedure call.
Each line is the ``repr`` output of a Python dictionary. For example (pretty
printed for display purposes, the original format is on one line)::

	{
		'name': 'person_insert',
		'args': {
			'per_id': 0,
			'per_firstname': 'Max',
			'per_lastname': 'Mustermann',
			per_created: 'sysdate'
		},
		'keys': ['per_id'],
		'sqls': ['per_created'],
	}

The keys in this dictionary have the following meaning:

	``name`` : string (required)
		The name of the procedure to be called.

	``args`` : dictionary (required)
		A dictionary with the names of the parameters as keys and the parameter
		values as values.

	``keys`` : list (optional)
		A list of parameter names that should be treated as keys. The value of the
		parameter is a integer or string identifer that is unique for each use of
		the key. On first use the parameter is used as an ``OUT`` parameter where
		the procedure will store the value of this key. On subsequent uses of this
		key (i.e. a key that has the same identifier) oradd will pass the value
		from the first use as a normal ``IN`` parameter. The ``keys`` key is
		optional, without it no parameter will be treated as a key.

	``sqls`` : list (optional)
		A list of parameter names that should be treated as SQL expressions.
		In the example above the parameter ``per_created`` will not be the string
		``"sysdate"``, but the result of the Oracle PL/SQL ``sysdate`` function.
		The ``sqls`` key is optional, without it no parameter will be treated as
		an SQL expression.


An oradd file in UL4ON format contains the same dictionaries, but not as a
Python repr output, but in UL4ON format. The UL4ON dump is *not* a list of
dictionaries, but simple concatenated dumps of each dictionary. When importing
this format ``oradd`` will simply read dumps from the file until the end of file
is reached.


Usage as a script
-----------------

``oradd.py`` has no external dependencies (except for :mod:`cx_Oracle`) and can
be used as a script for importing an oradd dump into the database. As a script
it supports the following command line options:

	``connectstring``
		An Oracle connectstring.

	``file``
		The name of the file from which the oradd dump is read. If ``file`` isn't
		specified the dump is read from ``stdin``.

	``-f``, ``--format``
		The format of the dump file: Either ``oradd`` (the default) or ``ul4on``.

	``-v``, ``--verbose``
		Gives different levels of output while data is imported to the database.
		Possible levels are: ``0`` (no output), ``1`` (a dot for each procedure
		call), ``2`` (like ``1``, plus a summary of which procedure has been
		called how often), ``3`` (detailed output for each procedure call, plus
		summary)
"""

# We're importing ``datetime``, so that it's available to ``eval()``
import sys, io, argparse, operator, collections, contextlib, datetime

import cx_Oracle


__docformat__ = "reStructuredText"


class Key(object):
	"""
	:class:`Key` instances are used to mark procedure values that are
	primary/foreign keys. On first use the parameter is used as an ``OUT``
	parameter and the procedure stores the value of the newly created primary key
	in this parameter. When a :class:`Key` object is used a second time its value
	will be passed to the procedure as normal ``IN`` parameters.
	"""

	seq = 0

	def __init__(self):
		self._value = self.__class__.seq
		self.__class__.seq += 1

	def value(self):
		return self._value


class SQL(object):
	"""
	An :class:`SQL` object can be used to specify an SQL expression as a
	procedure parameter instead of a fixed value (e.g. passing the current
	date (i.e. the date of the import) can be done with ``SQL("sysdate")``).
	"""

	def __init__(self, expression):
		self.expression = expression

	def value(self):
		return self.expression


def dump(name, **kwargs):
	"""
	Return the dump format for calling the procedure ``name`` with the parameters
	``kwargs``.

	``kwargs`` may contain :class:`Key` and :class`SQL` instances.
	"""
	keys = [key for (key, value) in kwargs.items() if isinstance(value, Key)]
	sqls = [key for (key, value) in kwargs.items() if isinstance(value, SQL)]
	for (key, value) in kwargs.items():
		if isinstance(value, (Key, SQL)):
			kwargs[key] = value.value()
	result = dict(name=name, args=kwargs)
	if keys:
		result["keys"] = keys
	if sqls:
		result["sqls"] = sqls
	return result


def dumps_oradd(name, **kwargs):
	"""
	Return the dump of a procedure call to the procedure named ``name`` with the
	parameters ``kwargs`` in oradd native format as a string.
	"""
	return "{!r}\n".format(dump(name, **kwargs))


def dump_oradd(stream, name, **kwargs):
	"""
	Dump a procedure call to the procedure named ``name`` with the parameters
	``kwargs`` into the output stream ``stream`` in oradd native format.
	"""
	stream.write(repr(dump(name, **kwargs)))
	stream.write("\n")


def dumps_ul4on(name, **kwargs):
	"""
	Return the dump of a procedure call to the procedure named ``name`` with the
	parameters ``kwargs`` in UL4ON format as a string.
	"""
	from ll import ul4on
	return ul4on.dumps(dump(name, **kwargs))


def dump_ul4on(stream, name, **kwargs):
	"""
	Dump a procedure call to the procedure named ``name`` with the parameters
	``kwargs`` into the output stream ``stream`` in UL4ON format.
	"""
	from ll import ul4on
	ul4on.dump(dump(name, **kwargs), stream)


def loads_oradd(string):
	"""
	Load an oradd dump in oradd native format from the string ``string``.

	This function is generator. It's output can be passed to :func:`importdata`.
	"""
	for line in string.splitlines():
		yield eval(line)


def load_oradd(stream):
	"""
	Load an oradd dump in oradd native format from the stream ``stream``.

	This function is generator. It's output can be passed to :func:`importdata`.
	"""
	for line in stream:
		yield eval(line)


def loads_ul4on(string):
	"""
	Load an oradd dump in UL4ON format from the string ``string``.

	This function is generator. It's output can be passed to :func:`importdata`.
	"""
	from ll import ul4on
	stream = io.StringIO(string)
	while True:
		try:
			yield ul4on.load(stream)
		except EOFError:
			break


def load_ul4on(stream):
	"""
	Load an oradd dump in UL4ON format from the stream ``stream``.

	This function is generator. It's output can be passed to :func:`importdata`.
	"""
	from ll import ul4on
	while True:
		try:
			yield ul4on.load(stream)
		except EOFError:
			break


def importdata(data, cursor):
	"""
	Import the data ``data`` produced by :func:`loads_oradd`, :func:`load_oradd`,
	:func:`loads_ulon` or :func:`load_ul4on` into the database. ``cursor`` must
	be a :mod:`cx_Oracle` cursor.

	:func:`importdata` is itself a generator. For each procedure call it produces
	the original data import, plus two additional keys: ``newkeyvalues`` is
	a dictionary containing the values of all keys that have been initialized
	in this procedure call. ``oldkeyvalues`` contains the values of all keys
	that have been reused for this procedure call.
	"""
	allkeys = {}
	for record in data:
		name = record["name"]
		args = record["args"]
		keys = set(record.get("keys", []))
		sqls = set(record.get("sqls", []))
		queryargvalues = {}
		queryargvars = {}
		for (argname, argvalue) in args.items():
			if argname in keys:
				queryargvalues[argname] = ":{}".format(argname)
				if argvalue in allkeys:
					queryargvars[argname] = allkeys[argvalue]
				else:
					queryargvars[argname] = cursor.var(int)
			elif argname in sqls:
				queryargvalues[argname] = argvalue
				# no value
			elif isinstance(argvalue, str) and len(argvalue) >= 4000:
				queryargvalues[argname] = ":{}".format(argname)
				var = cursor.var(cx_Oracle.CLOB)
				var.setvalue(0, argvalue)
				queryargvars[argname] = var
			else:
				queryargvalues[argname] = ":{}".format(argname)
				queryargvars[argname] = argvalue
		query = "begin {}({}); end;".format(name, ", ".join("{}=>{}".format(*argitem) for argitem in queryargvalues.items()))
		cursor.execute(query, queryargvars)

		newkeyvalues = {}
		oldkeyvalues = {}
		for (argname, argvalue) in args.items():
			if argname in keys:
				if argvalue in allkeys:
					oldkeyvalues[argname] = allkeys[argvalue]
				else:
					newkeyvalues[argname] = allkeys[argvalue] = queryargvars[argname].getvalue(0)
		record = record.copy()
		record["newkeyvalues"] = newkeyvalues
		record["oldkeyvalues"] = oldkeyvalues
		yield record


def main(args=None):
	p = argparse.ArgumentParser(description="Import an oradd dump to an Oracle database", epilog="For more info see http://www.livinglogic.de/Python/oradd/index.html")
	p.add_argument("connectstring", help="Oracle connect string")
	p.add_argument("file", nargs="?", help="Name of dump file (default: read from stdin)", type=argparse.FileType("r"), default=sys.stdin)
	p.add_argument("-f", "--format", dest="format", help="Format of the dumpfile ('oradd' or 'ul4on') (default %(default)s)", default="oradd", choices=("oradd", "ul4on"))
	p.add_argument("-v", "--verbose", dest="verbose", help="Give a progress report? (default %(default)s)", type=int, default=2, choices=(0, 1, 2, 3))

	args = p.parse_args(args)

	db = cx_Oracle.connect(args.connectstring)

	def formatdata(data):
		args = []
		newkeys = []
		for (argname, argvalue) in data["args"].items():
			if argname in data["newkeyvalues"]:
				args.append("{}=?".format(argname))
				newkeys.append("{}={}".format(argname, data["newkeyvalues"][argname]))
			elif argname in data["oldkeyvalues"]:
				args.append("{}={}".format(argname, data["oldkeyvalues"][argname]))
			else:
				args.append("{}={!r}".format(argname, argvalue))
		msg = "{}({})".format(data["name"], ", ".join(args))
		if newkeys:
			msg += " -> {}".format(", ".join(newkeys))
		return msg

	try:
		counts = collections.Counter()
		loader = dict(oradd=load_oradd, ul4on=load_ul4on)[args.format]
		for (i, data) in enumerate(importdata(loader(args.file), db.cursor()), start=1):
			if args.verbose >= 1:
				if args.verbose >= 3:
					print("#{}: {}".format(i, formatdata(data)))
				else:
					sys.stdout.write(".")
					sys.stdout.flush()
			counts[data["name"]] += 1
		db.commit()
	finally:
		if 1 <= args.verbose <= 2:
			print()

	if args.verbose >= 2:
		totalcount = sum(counts.values())
		l1 = len(str(totalcount))
		l2 = max(len(procname) for procname in counts)
		print()
		print("Summary")
		print("="*(l1+1+l2))
		print("{:>{}} procedure".format("#", l1))
		print("{} {}".format("-"*l1, "-"*l2))
		for (procname, count) in sorted(counts.items(), key=operator.itemgetter(1)):
			print("{:>{}} {}".format(count, l1, procname))
		print("{} {}".format("-"*l1, "-"*l2))
		print("{:>{}} (total calls)".format(totalcount, l1))


if __name__ == "__main__":
	sys.exit(main())



#! /usr/bin/env/python
# -*- coding: utf-8 -*-

## Copyright 1999-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


from ll.xist import xsc


def test_locationeq():
	l1 = xsc.Location(url="http://gurk.com", line=42, col=666)
	l2 = xsc.Location(url="http://gurk.com", line=42, col=666)
	l3 = xsc.Location(url="http://hurz.com", line=42, col=666)
	l4 = xsc.Location(url="http://gurk.com", line=43, col=666)
	l5 = xsc.Location(url="http://gurk.com", line=43, col=667)
	l6 = xsc.Location(url="http://gurk.com")
	assert l1 == l2
	assert l1 != l3
	assert l1 != l4
	assert l1 != l5
	assert l1 != l6


def test_locationoffset():
	l1 = xsc.Location(url="http://gurk.com", line=42, col=666)
	assert l1 == l1.offset(0)
	l2 = l1.offset(1)
	assert l1.url == l2.url
	assert l1.line+1 == l2.line



# -*- coding: utf-8 -*-

## Copyright 2009-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 2009-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


__docformat__ = "reStructuredText"

import os, collections, contextlib

from ll import url
from ll.xist import xsc, sims
from ll.xist.ns import doc


try:
	from docutils import core, nodes, utils, parsers, frontend
	from docutils.parsers.rst import roles
	from docutils.parsers.rst.languages import en
except ImportError:
	pass
else:
	class mod(nodes.literal):
		pass

	roles.register_generic_role("mod", mod)
	en.roles["mod"] = "mod"

	class class_(nodes.literal):
		pass

	roles.register_generic_role("class", class_)
	en.roles["class"] = "class"

	class func(nodes.literal):
		pass

	roles.register_generic_role("func", func)
	en.roles["func"] = "func"

	class meth(nodes.literal):
		pass

	roles.register_generic_role("meth", meth)
	en.roles["meth"] = "meth"

	class var(nodes.literal):
		pass

	roles.register_generic_role("var", var)
	en.roles["var"] = "var"

	class exc(nodes.literal):
		pass

	roles.register_generic_role("exc", exc)
	en.roles["exc"] = "exc"

	class attr(nodes.literal):
		pass

	roles.register_generic_role("attr", attr)
	en.roles["attr"] = "attr"

	class prop(nodes.literal):
		pass

	roles.register_generic_role("prop", prop)
	en.roles["prop"] = "prop"

	class option(nodes.literal):
		pass

	roles.register_generic_role("option", option)
	en.roles["option"] = "option"

	class const(nodes.literal):
		pass

	roles.register_generic_role("const", const)
	en.roles["const"] = "const"

	class file(nodes.literal):
		pass

	roles.register_generic_role("file", file)
	en.roles["file"] = "file"

	class dir(nodes.literal):
		pass

	roles.register_generic_role("dir", dir)
	en.roles["dir"] = "dir"

	class data(nodes.literal):
		pass

	roles.register_generic_role("data", data)
	en.roles["data"] = "data"


class DocTypeHTML(xsc.DocType):
	"""
	document type for ReST
	"""
	def __init__(self):
		xsc.DocType.__init__(self, 'document PUBLIC "+//IDN docutils.sourceforge.net//DTD Docutils Generic//EN//XML" "http://docutils.sourceforge.net/docs/ref/docutils.dtd"')


class refuriattrs(xsc.Attrs):
	class refuri(xsc.URLAttr):
		"External reference to a URI/URL"


class refidattrs(xsc.Attrs):
	class refid(xsc.TextAttr):
		"Internal reference to the `id` attribute of an element."


class backrefsattrs(xsc.Attrs):
	class backrefs(xsc.TextAttr):
		"Space-separated list of id references, for backlinks."


class refnameattrs(xsc.Attrs):
	class refname(xsc.TextAttr):
		"""
		Internal reference to the ``name`` attribute of an element. On a
		'target' element, 'refname' indicates an indirect target which may
		resolve to either an internal or external reference.
		"""


class referenceattrs(refuriattrs, refidattrs, refnameattrs):
	"Collected hyperlink reference attributes."


class anonymousattrs(xsc.Attrs):
	class anonymous(xsc.TextAttr):
		"Unnamed hyperlink"


class autoattrs(xsc.Attrs):
	class auto(xsc.TextAttr):
		"Auto-numbered footnote or title."


class alignhattrs(xsc.Attrs):
	class align(xsc.TextAttr):
		values = ("left", "center", "right")


class alignhvattrs(xsc.Attrs):
	class align(xsc.TextAttr):
		values = ("top", "middle", "bottom", "left", "center", "right")


xmlns = "http://docutils.sourceforge.net/ns"


class BaseElement(xsc.Element):
	xmlns = xmlns
	class Attrs(xsc.Element.Attrs):
		class ids(xsc.TextAttr):
			"``id`` is a unique identifier, typically assigned by the system."
		class names(xsc.TextAttr):
			"``name`` is an identifier assigned in the markup."
		class dupnames(xsc.TextAttr):
			"``dupname`` is the same as ``name``, used when it's a duplicate."
		class source(xsc.TextAttr):
			"``source`` is the name of the source of this document or fragment."
		class class_(xsc.TextAttr):
			"``class`` is used to transmit individuality information forward."
			xmlname = "class"


# Root Element
class document(BaseElement):
	class Attrs(BaseElement.Attrs):
		class title(xsc.TextAttr): pass

	def convert(self, converter):
		return self.content.convert(converter)


# Title Elements
class title(BaseElement):
	class Attrs(BaseElement.Attrs, refidattrs, autoattrs):
		pass

	def convert(self, converter):
		e = doc.h(self.content)
		return e.convert(converter)


class subtitle(BaseElement):
	pass


# Bibliographic Elements
class docinfo(BaseElement):
	pass


class info(BaseElement):
	pass


class author(BaseElement):
	pass


class authors(BaseElement):
	pass


class organization(BaseElement):
	pass


class address(BaseElement):
	pass


class contact(BaseElement):
	pass


class version(BaseElement):
	pass


class revision(BaseElement):
	pass


class status(BaseElement):
	pass


class date(BaseElement):
	pass


class copyright(BaseElement):
	pass


# Decoration Elements
class decoration(BaseElement):
	pass


class header(BaseElement):
	pass


class footer(BaseElement):
	pass


# Structural Elements
class section(BaseElement):
	def convert(self, converter):
		e = doc.section(self.content)
		return e.convert(converter)


class topic(BaseElement):
	pass


class sidebar(BaseElement):
	pass


class transition(BaseElement):
	model = sims.Empty()


# Body Elements
class paragraph(BaseElement):
	def convert(self, converter):
		e = doc.p(self.content)
		return e.convert(converter)


class compound(BaseElement):
	pass


class container(BaseElement):
	pass


class bullet_list(BaseElement):
	class Attrs(BaseElement.Attrs):
		class bullet(xsc.TextAttr): pass

	def convert(self, converter):
		e = doc.ul(self.content)
		return e.convert(converter)


class enumerated_list(BaseElement):
	class Attrs(BaseElement.Attrs):
		class enumtype(xsc.TextAttr):
			values = ("arabic", "loweraplha", "upperalpha", "lowerroman", "upperroman")
		class prefix(xsc.TextAttr): pass
		class suffix(xsc.TextAttr): pass
		class start(xsc.IntAttr): pass

	def convert(self, converter):
		e = doc.ol(self.content)
		return e.convert(converter)


class list_item(BaseElement):
	def convert(self, converter):
		e = doc.li(self.content)
		return e.convert(converter)


class definition_list(BaseElement):
	def convert(self, converter):
		e = doc.dl(self.content)
		return e.convert(converter)


class definition_list_item(BaseElement):
	def convert(self, converter):
		with xsc.build():
			with xsc.Frag() as e:
				if self[term] or self[classifier]:
					with doc.dt():
						+xsc.Frag(self[term])
						if self[classifier]:
							+xsc.Text(" (")
							+xsc.Frag(self[classifier]).withsep(", ")
							+xsc.Text(")")
				+xsc.Frag(self[definition])
		return e.convert(converter)


class term(BaseElement):
	def convert(self, converter):
		e = self.content
		return e.convert(converter)


class classifier(BaseElement):
	def convert(self, converter):
		e = self.content
		return e.convert(converter)


class definition(BaseElement):
	def convert(self, converter):
		e = doc.dd(self.content)
		return e.convert(converter)


class field_list(BaseElement):
	def convert(self, converter):
		e = doc.dl(self.content)
		return e.convert(converter)


class field(BaseElement):
	def convert(self, converter):
		e = self.content
		return e.convert(converter)


class field_name(BaseElement):
	def convert(self, converter):
		e = doc.dt(self.content)
		return e.convert(converter)


class field_body(BaseElement):
	def convert(self, converter):
		e = doc.dd(self.content)
		return e.convert(converter)


class option_list(BaseElement):
	def convert(self, converter):
		e = doc.dl(self.content)
		return e.convert(converter)


class option_list_item(BaseElement):
	def convert(self, converter):
		e = self.content
		return e.convert(converter)


class option_group(BaseElement):
	def convert(self, converter):
		e = doc.dt()
		for o in self[option]:
			if e:
				e.append(", ")
			e2 = doc.lit(doc.option(o[option_string][0].content))
			for oa in o[option_argument]:
				e2.append("=", oa.content)
			e.append(e2)
		return e.convert(converter)


class option_string(BaseElement):
	model = sims.NoElements()


class option_argument(BaseElement):
	model = sims.NoElements()
	class Attrs(BaseElement.Attrs):
		class delimiter(xsc.TextAttr): pass


class description(BaseElement):
	def convert(self, converter):
		e = doc.dd(self.content)
		return e.convert(converter)


class literal_block(BaseElement):
	def convert(self, converter):
		e = doc.litblock(self.content)
		return e.convert(converter)


class line_block(BaseElement):
	pass


class line(BaseElement):
	pass


class block_quote(BaseElement):
	pass


class attribution(BaseElement):
	pass


class doctest_block(BaseElement):
	pass


class attention(BaseElement):
	pass


class caution(BaseElement):
	pass


class danger(BaseElement):
	pass


class error(BaseElement):
	pass


class hint(BaseElement):
	pass


class important(BaseElement):
	pass


class note(BaseElement):
	pass


class tip(BaseElement):
	pass


class warning(BaseElement):
	pass


class admonition(BaseElement):
	pass


class footnote(BaseElement):
	class Attrs(BaseElement.Attrs, backrefsattrs, autoattrs):
		pass


class citation(BaseElement):
	class Attrs(BaseElement.Attrs, backrefsattrs):
		pass


class label(BaseElement):
	model = sims.NoElements()
	pass


class rubric(BaseElement):
	pass


class target(BaseElement):
	class Attrs(BaseElement.Attrs, referenceattrs, anonymousattrs):
		pass

	def convert(self, converter):
		return xsc.Null


class substitution_definition(BaseElement):
	class Attrs(BaseElement.Attrs):
		class ltrim(xsc.TextAttr): pass
		class rtrim(xsc.TextAttr): pass


class comment(BaseElement):
	model = sims.NoElements()
	pass


class pending(BaseElement):
	model = sims.Empty()
	pass


class figure(BaseElement):
	class Attrs(BaseElement.Attrs, alignhattrs):
		class width(xsc.TextAttr): pass


class image(BaseElement):
	model = sims.Empty()
	class Attrs(BaseElement.Attrs, alignhvattrs):
		class uri(xsc.URLAttr): required = True
		class alt(xsc.TextAttr): pass
		class height(xsc.TextAttr): pass
		class width(xsc.TextAttr): pass
		class scale(xsc.TextAttr): pass


class caption(BaseElement):
	pass


class legend(BaseElement):
	pass


# Table elements: table, tgroup, colspec, thead, tbody, row, entry.


class system_message(BaseElement):
	"Used to record processing information"
	class Attrs(BaseElement.Attrs, backrefsattrs):
		class level(xsc.TextAttr): pass
		class line(xsc.TextAttr): pass
		class type(xsc.TextAttr): pass

	def convert(self, converter):
		# A warning has already been issued by docutils, we don't have to do anything
		return xsc.Null


class raw(BaseElement):
	class Attrs(BaseElement.Attrs):
		class format(xsc.TextAttr): pass


# Inline elements
class emphasis(BaseElement):
	def convert(self, converter):
		e = doc.em(self.content)
		return e.convert(converter)

class strong(BaseElement):
	def convert(self, converter):
		e = doc.em(self.content)
		return e.convert(converter)


class literal(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.lit(self.content)
		return e.convert(converter)


class meth(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.meth(self.content)
		return e.convert(converter)


class mod(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.mod(self.content)
		return e.convert(converter)


class class_(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.class_(self.content)
		return e.convert(converter)


class func(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.func(self.content)
		return e.convert(converter)


class meth(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.meth(self.content)
		return e.convert(converter)


class var(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.var(self.content)
		return e.convert(converter)


class exc(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.exc(self.content)
		return e.convert(converter)


class attr(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.attr(self.content)
		return e.convert(converter)


class prop(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.prop(self.content)
		return e.convert(converter)


class option(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.option(self.content)
		return e.convert(converter)


class const(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.const(self.content)
		return e.convert(converter)


class file(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.file(self.content)
		return e.convert(converter)


class dir(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.dir(self.content)
		return e.convert(converter)


class data(BaseElement):
	model = sims.NoElements()

	def convert(self, converter):
		e = doc.data(self.content)
		return e.convert(converter)


class reference(BaseElement):
	class Attrs(BaseElement.Attrs, referenceattrs, anonymousattrs):
		class name(xsc.TextAttr): pass

	def convert(self, converter):
		e = doc.a(self.content, href=self.attrs.refuri)
		return e.convert(converter)


class footnote_reference(BaseElement):
	model = sims.NoElements()
	class Attrs(BaseElement.Attrs, refidattrs, refnameattrs, autoattrs):
		pass


class citation_reference(BaseElement):
	model = sims.NoElements()
	class Attrs(BaseElement.Attrs, refidattrs, refnameattrs):
		pass


class substitution_reference(BaseElement):
	model = sims.NoElements()
	class Attrs(BaseElement.Attrs, refnameattrs):
		pass


class title_reference(BaseElement):
	pass


class abbreviation(BaseElement):
	pass


class acronym(BaseElement):
	pass


class superscript(BaseElement):
	pass


class subscript(BaseElement):
	pass


class inline(BaseElement):
	pass


class inline(BaseElement):
	class Attrs(BaseElement.Attrs, refidattrs):
		pass


class problematic(BaseElement):
	class Attrs(BaseElement.Attrs, refidattrs):
		pass


class generated(BaseElement):
	pass


PE_inline_elements = (
	emphasis,
	strong,
	literal,
	reference,
	footnote_reference,
	citation_reference,
	substitution_reference,
	title_reference,
	abbreviation,
	acronym,
	subscript,
	superscript,
	inline,
	problematic,
	generated,
	target,
	image,
	raw,
)

PE_body_elements = (
	paragraph,
	compound,
	container,
	literal_block,
	doctest_block,
	line_block,
	block_quote,
#	table,
	figure,
	image,
	footnote,
	citation,
	rubric,
	bullet_list,
	enumerated_list,
	definition_list,
	field_list,
	option_list,
	attention,
	caution,
	danger,
	error,
	hint,
	important,
	note,
	tip,
	warning,
	admonition,
	reference,
	target,
	substitution_definition,
	comment,
	pending,
	system_message,
	raw,
)

PE_section_elements = (section,)

PE_bibliographic_elements = (
	author,
	authors,
	organization,
	address,
	contact,
	version,
	revision,
	status,
	date,
	copyright,
	field,
)

PE_structure_model = (topic, sidebar, transition) + PE_section_elements

document.model = sims.Elements(*(title, subtitle, decoration, docinfo, transition) + PE_structure_model)
title.model = sims.ElementsOrText(*PE_inline_elements)
subtitle.model = sims.ElementsOrText(*PE_inline_elements)
docinfo.model = sims.Elements(*PE_bibliographic_elements)
author.model = sims.ElementsOrText(*PE_inline_elements)
authors.model = sims.Elements(author, organization, address, contact)
organization.model = sims.ElementsOrText(*PE_inline_elements)
address.model = sims.ElementsOrText(*PE_inline_elements)
contact.model = sims.ElementsOrText(*PE_inline_elements)
version.model = sims.ElementsOrText(*PE_inline_elements)
revision.model = sims.ElementsOrText(*PE_inline_elements)
status.model = sims.ElementsOrText(*PE_inline_elements)
date.model = sims.ElementsOrText(*PE_inline_elements)
copyright.model = sims.ElementsOrText(*PE_inline_elements)
decoration.model = sims.Elements(header, footer)
header.model = sims.Elements(*PE_body_elements)
footer.model = sims.Elements(*PE_body_elements)
section.model = sims.Elements(*((title, subtitle, info, decoration) + PE_structure_model))
topic.model = sims.Elements(*((title,) + PE_body_elements))
sidebar.model = sims.Elements(*((title, subtitle, topic) + PE_body_elements))
paragraph.model = sims.ElementsOrText(*PE_inline_elements)
compound.model = sims.Elements(*PE_body_elements)
container.model = sims.Elements(*PE_body_elements)
bullet_list.model = sims.Elements(list_item)
enumerated_list.model = sims.Elements(list_item)
list_item.model = sims.Elements(*PE_body_elements)
definition_list.model = sims.Elements(definition_list_item)
definition_list_item.model = sims.Elements(term, classifier, definition)
term.model = sims.ElementsOrText(*PE_inline_elements)
classifier.model = sims.ElementsOrText(*PE_inline_elements)
definition.model = sims.Elements(*PE_body_elements)
field_list.model = sims.Elements(field)
field.model = sims.Elements(field_name, field_body)
field_name.model = sims.ElementsOrText(*PE_inline_elements)
field_body.model = sims.Elements(*PE_body_elements)
option_list.model = sims.Elements(option_list_item)
option_list_item.model = sims.Elements(option_group, description)
option_group.model = sims.Elements(option)
option.model = sims.Elements(option_string, option_argument)
description.model = sims.Elements(*PE_body_elements)
literal_block.model = sims.ElementsOrText(*PE_inline_elements)
line_block.model = sims.Elements(line, line_block)
line.model = sims.ElementsOrText(*PE_inline_elements)
block_quote.model = sims.Elements(*(PE_body_elements + (attribution,)))
attribution.model = sims.ElementsOrText(*PE_inline_elements)
doctest_block.model = sims.ElementsOrText(*PE_inline_elements)
attention.model = sims.Elements(*PE_body_elements)
caution.model = sims.Elements(*PE_body_elements)
danger.model = sims.Elements(*PE_body_elements)
error.model = sims.Elements(*PE_body_elements)
hint.model = sims.Elements(*PE_body_elements)
important.model = sims.Elements(*PE_body_elements)
note.model = sims.Elements(*PE_body_elements)
tip.model = sims.Elements(*PE_body_elements)
warning.model = sims.Elements(*PE_body_elements)
admonition.model = sims.Elements(*(PE_body_elements + (title,)))
citation.model = sims.Elements(*(PE_body_elements + (label,)))
rubric.model = sims.ElementsOrText(*PE_inline_elements)
target.model = sims.ElementsOrText(*PE_inline_elements)
substitution_definition.model = sims.ElementsOrText(*PE_inline_elements)
figure.model = sims.Elements(image, caption, legend)
caption.model = sims.ElementsOrText(*PE_inline_elements)
legend.model = sims.Elements(*PE_body_elements)
system_message.model = sims.Elements(*PE_body_elements)
raw.model = sims.ElementsOrText(*PE_inline_elements)
emphasis.model = sims.ElementsOrText(*PE_inline_elements)
strong.model = sims.ElementsOrText(*PE_inline_elements)
reference.model = sims.ElementsOrText(*PE_inline_elements)
substitution_reference.model = sims.ElementsOrText(*PE_inline_elements)
title_reference.model = sims.ElementsOrText(*PE_inline_elements)
abbreviation.model = sims.ElementsOrText(*PE_inline_elements)
acronym.model = sims.ElementsOrText(*PE_inline_elements)
superscript.model = sims.ElementsOrText(*PE_inline_elements)
subscript.model = sims.ElementsOrText(*PE_inline_elements)
inline.model = sims.ElementsOrText(*PE_inline_elements)
problematic.model = sims.ElementsOrText(*PE_inline_elements)
generated.model = sims.ElementsOrText(*PE_inline_elements)


class ReSTConversionWarning(Warning):
	pass


class ReSTConverter(object):
	def __init__(self):
		from ll.xist.ns import doc, abbr
		self.namedrefs = collections.defaultdict(list)
		self.unnamedrefs = []
		self.doc = doc
		self.abbr = abbr

	def convert(self, node):
		if isinstance(node, nodes.document):
			return xsc.Frag(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.Text):
			return xsc.Text(node.astext())
		elif isinstance(node, nodes.problematic):
			# We don't do anything about this
			return xsc.Frag(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.section):
			return self.doc.section(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.title):
			return self.doc.h(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.paragraph):
			return self.doc.p(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.bullet_list):
			return self.doc.ul(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.list_item):
			return self.doc.li(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.definition_list):
			return self.doc.dl(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.definition_list_item):
			return xsc.Frag(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.term):
			return self.doc.dt(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.definition):
			return self.doc.dd(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.literal_block):
			return self.doc.prog(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.literal):
			return self.doc.lit(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.emphasis):
			return self.doc.em(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.substitution_reference):
			try:
				return getattr(self.abbr, node.attributes["refname"].lower())()
			except AttributeError:
				return xsc.Frag(self.convert(child) for child in node.children)
		elif isinstance(node, nodes.reference):
			e = self.doc.a(self.convert(child) for child in node.children)
			if "anonymous" in node.attributes:
				self.unnamedrefs.append(e)
			else:
				self.namedrefs[node.attributes["refname"]].append(e)
			return e
		elif isinstance(node, nodes.target):
			uri = node.attributes["refuri"]
			if "anonymous" in node.attributes:
				# Set the link on the first unnamed reference
				self.unnamedrefs[0].attrs.href = uri
				del self.unnamedrefs[0] # done => remove it
			else:
				for name in node.attributes["names"]:
					try:
						es = self.namedrefs[name]
					except KeyError:
						pass
					else:
						for e in es:
							e.attrs.href = uri
						del self.namedrefs[name]
			return xsc.Null
		elif isinstance(node, nodes.system_message):
			warnings.warn(ReSTConversionWarning(str(node)))
			return xsc.Null # ignore system messages
		else:
			raise TypeError("can't handle {!r}".format(node.__class__))


def fromstring(string, base=None, **options):
	realoptions = dict(tab_width=3)
	realoptions.update(options)
	doc = core.publish_doctree(string, source_path=base, settings_overrides=realoptions) # This requires docutils
	elements = globals()

	def toxist(node):
		if isinstance(node, nodes.Text):
			return xsc.Text(str(node.astext()))
		else:
			e = elements[node.__class__.__name__](toxist(child) for child in node.children)
			e.startloc = xsc.Location(node.source, node.line)
			for (attrkey, attrvalue) in node.attlist():
				if attrkey in e.Attrs:
					if isinstance(attrvalue, list):
						attrvalue = " ".join(attrvalue)
					e.attrs[attrkey] = attrvalue
			return e

	return toxist(doc)


def fromstream(stream, base=None, **options):
	return fromstring(stream.read(), base, **options)


def fromfile(filename, base=None, **options):
	filename = os.path.expanduser(filename)
	if base is None:
		base = filename
	with contextlib.closing(open(filename, "rb")) as stream:
		return fromstream(stream, base, **options)



#! /usr/bin/env/python
# -*- coding: utf-8 -*-

## Copyright 1999-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license

import pickle

from ll.xist import xsc
from ll.xist.ns import xml, html, chars, abbr, php


def test_pickle():
	e = xsc.Frag(
		xml.XML(),
		html.DocTypeXHTML10transitional(),
		xsc.Comment("foo"),
		html.html(xml.Attrs(lang="de"), lang="de"),
		php.expression("$foo"),
		chars.nbsp(),
		abbr.xml(),
	)
	e.append(e[3])
	e2 = pickle.loads(pickle.dumps(e, 2))
	assert e == e2
	assert e2[3] is e2[-1]



#! /usr/bin/env/python
# -*- coding: utf-8 -*-

## Copyright 2005-2013 by LivingLogic AG, Bayreuth/Germany.
## Copyright 2005-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


import sys, re

from ll import misc

import pytest


# The following includes \x00 in addition to those characters defined in
# http://www.w3.org/TR/2004/REC-xml11-20040204/#NT-RestrictedChar
restrictedchars = re.compile("[\x00-\x08\x0B-\x0C\x0E-\x1F\x7F-\x84\x86-\x9F]")


escape_input = "".join(chr(i) for i in range(1000)) + "".join(chr(i) for i in range(sys.maxunicode-10, sys.maxunicode+1))


def test_xmlescape():
	for input in (escape_input, escape_input.encode("iso-8859-1", "ignore")):
		escape_output = []
		for c in escape_input:
			if c=="&":
				escape_output.append("&amp;")
			elif c=="<":
				escape_output.append("&lt;")
			elif c==">":
				escape_output.append("&gt;")
			elif c=='"':
				escape_output.append("&quot;")
			elif c=="'":
				escape_output.append("&#39;")
			elif restrictedchars.match(c) is not None:
				escape_output.append("&#{};".format(ord(c)))
			else:
				escape_output.append(c)
		escape_output = "".join(escape_output)
		assert misc.xmlescape(escape_input) == escape_output


def test_xmlescape_text():
	for input in (escape_input, escape_input.encode("iso-8859-1", "ignore")):
		escape_output = []
		for c in escape_input:
			if c=="&":
				escape_output.append("&amp;")
			elif c=="<":
				escape_output.append("&lt;")
			elif c==">":
				escape_output.append("&gt;")
			elif restrictedchars.match(c) is not None:
				escape_output.append("&#{};".format(ord(c)))
			else:
				escape_output.append(c)
		escape_output = "".join(escape_output)
		assert misc.xmlescape_text(escape_input) == escape_output


def test_xmlescape_attr():
	for input in (escape_input, escape_input.encode("iso-8859-1", "ignore")):
		escape_output = []
		for c in escape_input:
			if c=="&":
				escape_output.append("&amp;")
			elif c=="<":
				escape_output.append("&lt;")
			elif c==">":
				escape_output.append("&gt;")
			elif c=='"':
				escape_output.append("&quot;")
			elif restrictedchars.match(c) is not None:
				escape_output.append("&#{};".format(ord(c)))
			else:
				escape_output.append(c)
		escape_output = "".join(escape_output)
		assert misc.xmlescape_attr(escape_input) == escape_output


def test_item():
	def err(n):
		yield from range(n)
		raise SyntaxError

	e = iter(range(10))
	assert misc.item(e, 0) == 0
	assert misc.item(e, 0) == 1
	assert misc.item(e, -1) == 9
	with pytest.raises(IndexError):
		misc.item(e, -1)
	assert misc.item(e, -1, 42) == 42

	e = iter(range(10))
	assert misc.item(e, 4) == 4

	e = iter(range(10))
	with pytest.raises(IndexError):
		misc.item(e, 10)

	e = iter(range(10))
	assert misc.item(e, 10, 42) == 42

	e = iter(range(10))
	assert misc.item(e, -1) == 9

	e = iter(range(10))
	assert misc.item(e, -10) == 0

	e = iter(range(10))
	with pytest.raises(IndexError):
		misc.item(e, -11)

	e = iter(range(10))
	assert misc.item(e, -11, 42) == 42

	iterable = [17, 23, 37]

	# Wrong arguments
	with pytest.raises(TypeError):
		misc.item()
	with pytest.raises(TypeError):
		misc.item([])
	with pytest.raises(TypeError):
		misc.item(42, 42)

	# Non-negative index
	assert misc.item(iterable, 0), 17
	assert misc.item(iterable, 2), 37
	with pytest.raises(IndexError):
		misc.item(iterable, 3)
	assert misc.item(iterable, 3, 42), 42
	assert misc.item(err(10), 9), 9
	with pytest.raises(SyntaxError):
		misc.item(err(10), 10)

	# Negative index
	assert misc.item(iterable, -1), 37
	assert misc.item(iterable, -3), 17
	with pytest.raises(IndexError):
		misc.item(iterable, -4)
	assert misc.item(iterable, -4, 42), 42
	# iterator is always exhausted
	with pytest.raises(SyntaxError):
		misc.item(err(10), -1)


def test_first():
	e = iter(range(10))
	assert misc.first(e) == 0
	assert misc.first(e) == 1

	e = iter([])
	with pytest.raises(IndexError):
		misc.first(e)

	e = iter([])
	assert misc.first(e, 42) == 42


def test_last():
	e = iter(range(10))
	assert misc.last(e) == 9
	with pytest.raises(IndexError):
		misc.last(e)

	e = iter([])
	with pytest.raises(IndexError):
		misc.last(e)

	e = iter([])
	assert misc.last(e, 42) == 42


def test_count():
	e = iter(range(10))
	assert misc.count(e) == 10
	assert misc.count(e) == 0

	e = iter([])
	assert misc.count(e) == 0


def test_iterator_bool():
	e = misc.Iterator(iter(range(10)))
	assert e

	e = misc.Iterator(iter([]))
	assert not e


def test_iterator_next():
	e = misc.Iterator(iter(range(2)))
	assert next(e) == 0
	assert next(e) == 1
	with pytest.raises(StopIteration):
		next(e)


def test_iterator_getitem():
	e = misc.Iterator(iter(range(10)))
	assert e[0] == 0
	assert e[0] == 1
	assert e[-1] == 9
	with pytest.raises(IndexError):
		e[-1]


def test_pool():
	pool = misc.Pool(misc)
	assert pool.Pool is pool["Pool"] is misc.Pool


def test_module():
	m = misc.module("a = 42")
	assert m.a == 42
	assert m.__name__ == "unnamed"
	assert m.__file__ == "unnamed.py"

	m = misc.module("a = 42", "/Users/walter/test.py")
	assert m.a == 42
	assert m.__name__ == "test"
	assert m.__file__ == "/Users/walter/test.py"

	m = misc.module("a = 42", "/Users/walter/test.py", "pest")
	assert m.a == 42
	assert m.__name__ == "pest"
	assert m.__file__ == "/Users/walter/test.py"


def test_itersplitat():
	assert tuple(misc.itersplitat("20090609172345", (4, 6, 8, 10, 12))) == ("2009", "06", "09", "17", "23", "45")
	assert tuple(misc.itersplitat("200906091723", (4, 6, 8, 10, 12))) == ("2009", "06", "09", "17", "23")
	assert tuple(misc.itersplitat("20090609172345", (-10, -8, -6, -4, -2))) == ("2009", "06", "09", "17", "23", "45")


def test_gzip():
	assert misc.gunzip(misc.gzip(b"gurk", 0)) == b"gurk"
	assert misc.gunzip(misc.gzip(b"gurk", 9)) == b"gurk"


def test_jsmin():
	assert misc.jsmin("gurk \t = \t 42;") == "gurk=42;"


def test_notimplemented():
	class Bad(object):
		@misc.notimplemented
		def bad(self):
			pass

	with pytest.raises(NotImplementedError):
		Bad().bad()


def test_javaexpr():
	# None
	assert "null" == misc.javaexpr(None)
	# bool
	assert "true" == misc.javaexpr(True)
	assert "false" == misc.javaexpr(False)
	# int
	assert "42" == misc.javaexpr(42)
	assert str((1<<31)-1) == misc.javaexpr((1<<31)-1)
	assert str(1<<31)+"L" == misc.javaexpr(1<<31)
	assert str(-(1<<31)) == misc.javaexpr(-(1<<31))
	assert str(-(1<<31)-1)+"L" == misc.javaexpr(-(1<<31)-1)
	assert str((1<<63)-1) + "L" == misc.javaexpr((1<<63)-1)
	assert 'new java.math.BigInteger("{}")'.format(1<<64) == misc.javaexpr(1<<64)
	# float
	assert "42.5" == misc.javaexpr(42.5)
	assert "1e+20" == misc.javaexpr(1e20)
	# string
	assert '""' == misc.javaexpr("")
	assert '"abc"' == misc.javaexpr("abc")
	assert '"\'"' == misc.javaexpr("'")
	assert '"\\n"' == misc.javaexpr("\n")
	assert '"\\r"' == misc.javaexpr("\r")
	assert '"\\t"' == misc.javaexpr("\t")
	assert '"\\f"' == misc.javaexpr("\f")
	assert '"\\b"' == misc.javaexpr("\b")
	assert '"\\""' == misc.javaexpr('"')
	assert '"\\u0000"' == misc.javaexpr("\x00")
	assert '"\\u00ff"' == misc.javaexpr("\xff")
	assert '"\\u20ac"' == misc.javaexpr("\u20ac")
	# list
	assert "java.util.Arrays.asList()" == misc.javaexpr(())
	assert "java.util.Arrays.asList(1, 2, 3)" == misc.javaexpr([1, 2, 3])
	# dict
	assert "com.livinglogic.utils.MapUtils.makeMap()" == misc.javaexpr({})
	assert "com.livinglogic.utils.MapUtils.makeMap(1, 2)" == misc.javaexpr({1: 2})


def test_sysinfo():
	# At least make sure that we don't produce exceptions by accessing each attribute once
	misc.sysinfo.host_name
	misc.sysinfo.host_fqdn
	misc.sysinfo.host_ip
	misc.sysinfo.host_sysname
	misc.sysinfo.host_nodename
	misc.sysinfo.host_release
	misc.sysinfo.host_version
	misc.sysinfo.host_machine
	misc.sysinfo.python_executable
	misc.sysinfo.python_version
	misc.sysinfo.pid
	misc.sysinfo.scriptname
	misc.sysinfo.shortscriptname


def test_prettycsv():
	assert "".join(misc.prettycsv([["a", "b", "c"], ["abc", "defg", "hijkl"]])) == "a     b      c\nabc   defg   hijkl\n"
	assert "".join(misc.prettycsv([["a", "b "], ["abc", "def"]])) == "a     b\nabc   def\n"
	assert "".join(misc.prettycsv([["a"], ["abc", "def"]])) == "a\nabc   def\n"
	assert "".join(misc.prettycsv([["a", "b"], ["abc", "def"]], "..")) == "a  ..b\nabc..def\n"



#! /usr/bin/env python
# -*- coding: utf-8 -*-

## Copyright 1999-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


"""
Purpose
-------

``tld2xsc`` is a script that converts a JSP Tag Library Descriptor XML file
into a skeleton XIST namespace module. The tld file is read from stdin and
the namespace module is printed to stdout.


Options
-------

``tld2xsc`` supports the following options:

	``-s``, ``--shareattrs`` : ``none``, ``dupes``, ``all``
		Should attributes be shared among the elements? ``none`` means that each
		element will have its own standalone :class:`Attrs` class directly derived
		from :class:`ll.xist.Elements.Attrs`. For ``dupes`` each attribute that is
		used by more than one element will be moved into its own :class:`Attrs`
		class. For ``all`` this will be done for all attributes.

	``-m``, ``--model`` : ``no``, ``simple``, ``fullall``, ``fullonce``
		Add model information to the namespace. ``no`` doesn't add any model
		information. ``simple`` only adds ``model = False`` or ``model = True``
		(i.e. only the information whether the element must be empty or not).
		``fullall`` adds a :mod:`ll.xist.sims` model object to each element class.
		``fullonce`` adds full model information to, but reuses model objects for
		elements which have the same model.
"""

__docformat__ = "reStructuredText"


import sys, argparse

from ll import misc, url
from ll.xist import xsc, xfind, parse
from ll.xist.ns import tld


__docformat__ = "reStructuredText"


def makexnd(stream, encoding=None, shareattrs="dupes", model="simple"):
	# :var:`stream` can be a stream, an :class:`URL` or ``str``/``bytes``
	encoding = None
	if isinstance(stream, str):
		encoding = "utf-8"
		stream = stream.encode(encoding)
	node = parse.tree(stream, parse.Expat(encoding=encoding), parse.NS(tld), parse.Node())

	# get and convert the taglib object
	xnd = node.walknodes(tld.taglib)[0].asxnd(model=model)

	if shareattrs=="dupes":
		xnd.shareattrs(False)
	elif shareattrs=="all":
		xnd.shareattrs(True)
	return xnd


def main(args=None):
	p = argparse.ArgumentParser(description="Convert JSP Tag Library Descriptor XML file (on stdin) to XIST namespace (on stdout)", epilog="For more info see http://www.livinglogic.de/Python/xist/scripts/tld2xsc.html")
	p.add_argument("-s", "--shareattrs", dest="shareattrs", help="Should identical attributes be shared among elements? (default %(default)s)", choices=("none", "dupes", "all"), default="dupes")
	p.add_argument("-m", "--model", dest="model", help="Add sims information to the namespace (default %(default)s)", choices=("none", "simple", "fullall", "fullonce"), default="simple")

	args = p.parse_args(args)
	print(makexnd(sys.stdin, args.shareattrs, model=args.model))


if __name__ == "__main__":
	sys.exit(main())



# -*- coding: utf-8 -*-

## Copyright 2005-2013 by LivingLogic AG, Bayreuth/Germany
## Copyright 2005-2013 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


"""
This module is an XIST namespace for Kid__ files.

__ http://kid.lesscode.org/
"""


from ll.xist import xsc, sims


__docformat__ = "reStructuredText"


xmlns = "http://purl.org/kid/ns#"


class Attrs(xsc.Attrs):
	"""
	Global attributes.
	"""
	class for_(xsc.TextAttr):
		"""
		The :class:`for_` attribute may appear on any element to signify that
		the element should be processed multiple times, once for each value in
		the sequence specified.
		"""
		xmlns = xmlns
		xmlname = "for"

	class if_(xsc.TextAttr):
		"""
		The :class:`if_` attribute may appear on any element to signify that
		the element and its decendant items should be output only if the boolean
		expression specified evaluates to true in Python.
		"""
		xmlns = xmlns
		xmlname = "if"

	class content(xsc.TextAttr):
		"""
		This attribute may appear on any element to signify that the decendant
		items of the element are to be replaced with the result of evaluating
		the attribute content as a Python expression.
		"""
		xmlns = xmlns

	class replace(xsc.TextAttr):
		"""
		:class:`replace` is shorthand for specifying a :class:`content` and a
		``strip="True"`` on the same element.
		"""
		xmlns = xmlns

	class strip(xsc.TextAttr):
		"""
		The :class:`strip` attribute may apppear on any element to signify that
		the containing element should not be output.
		"""
		xmlns = xmlns

	class attrs(xsc.TextAttr):
		"""
		The :class:`attrs` attribute may appear on any element to specify a set
		of attributes that should be set on the element when it is processed.
		"""
		xmlns = xmlns

	class def_(xsc.TextAttr):
		"""
		The :class:`def_` attribute may appear on any element to create a
		"Named Template Function".
		"""
		xmlns = xmlns
		xmlname = "def"

	class match(xsc.TextAttr):
		"""
		The :class:`match` attribute may appear on any element to create a
		"Match Template".
		"""
		xmlns = xmlns

	class extends(xsc.TextAttr):
		"""
		The :class:`extends` attribute may appear on the root element to specify
		that the template should inherit the Named Template Functions and Match
		Templates defined in another template (or set of templates).
		"""
		xmlns = xmlns


class python(xsc.ProcInst):
	"""
	The :class:`python` processing instruction contains Python code.
	"""
